/*
 * drivers/staging/gs/provision.c
 *
 * Driver for the provison.
 *
 * Copyright (C) 2012, 2013 GrandStream Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/rslib.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/types.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/semaphore.h>
#include <linux/io.h>
#include <linux/slab.h>
#include <linux/completion.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <../fs/proc/internal.h>

#include <mach/gs_type.h>
#include "mmc_rw.h"

static prov_t provs[] = {
	{SEG_PRELOADER_STR, },
	{SEG_LK_STR, },
	{SEG_KERNEL_A_STR, },
	{SEG_SYSTEM_A_STR, },
	{SEG_KERNEL_B_STR, },
	{SEG_SYSTEM_B_STR, },
	{SEG_RECOVERY_STR },
};

static int provision_proc_show_status(struct seq_file *seq, void *v)
{
	int ret = 0, i = 0;

	if (ret == 0)
	ret = seq_printf(seq, "%-16s: %lu\n", "Start Sector", (unsigned long)IMGHDR_SECTOR_OFFSET);
	if (ret == 0)
	ret = seq_printf(seq, "%-16s: %lu\n", "Sectors", (unsigned long)IMGHDR_SECTOR_SIZE);

	for (i = 0; i < ARRAY_SIZE(provs); i++) {
		ret = seq_printf(seq, "%-16s: %lu\n", provs[i].name, provs[i].start);;
	}

	return ret;
}

static int provision_proc_show_mmc(struct seq_file *m, void *v)
{
	int len;
	int rval;

	prov_t *prov = (prov_t *)m->private;
	seg_header_t *hdr;
	sector_t start_sector;
	sector_t i;
	int j = 0;
	char *data = NULL;

	hdr = (seg_header_t *)vmalloc(PAGE_SIZE);
	if (!hdr)
		return (-ENOMEM);

	len = sizeof(seg_header_t);
	memset(hdr, 0, sizeof(seg_header_t));

	start_sector = prov->start;

	for (i = 0; i < IMGHDR_SECTOR_SIZE/ARRAY_SIZE(provs); i++) {
		rval = emmc_read(start_sector + i, 1, (char *)hdr, 512);
		if (0 == rval)
			break;
		printk("emmc_read image header sector(%lu) failed, try next\n", i);
	}
	/*if (hdr->signature != GS_PROD_SIGNATURE || rval != 0)
		memset(hdr, 0x0, sizeof(seg_header_t));*/

	data = (char *)hdr;
	for (j = 0; j < len; j++) {
		seq_printf(m, "%c", *data);
		data++;
	}

	vfree(hdr);
	return 0;
}


static ssize_t
provision_proc_write_mmc(struct file *file,
	const char __user *buffer, size_t count, loff_t *pos)

{
	int rval = 0;
	prov_t *prov = PDE_DATA(file_inode(file));
	seg_header_t *hdr;
	sector_t off;
	sector_t start;

	if (count > sizeof(seg_header_t))
		return (-EINVAL);

	hdr = (seg_header_t *)vmalloc(PAGE_SIZE);
	if (!hdr)
		return (-ENOMEM);

	memset(hdr, 0, sizeof(seg_header_t));
	if (copy_from_user(hdr, buffer, count)) {
		vfree(hdr);
		return (-EFAULT);
	}

	start = prov->start;
	for (off = 0; off < IMGHDR_SECTOR_SIZE/ARRAY_SIZE(provs); off++) {
		rval = emmc_write(start + off, 1, (char *)hdr, 512);
		if (0 == rval)
			break;
		printk("emmc_write image header sector(%lu) failed, try next\n", off);
	}

	vfree(hdr);

	return count;
}

static int provision_proc_show_updflag(struct seq_file *seq, void *v)
{
	int rval;
	upd_flag_t *upd_flag;
	sector_t off;
	int j = 0, len = 0;
	char *data = NULL;

	upd_flag = (upd_flag_t *)vmalloc(PAGE_SIZE);
	if (!upd_flag)
		return (-ENOMEM);

	len = sizeof(upd_flag_t);
	memset(upd_flag, 0, sizeof(upd_flag_t));

	for (off = 0; off < UPDFLAG_SECTOR_SIZE; off++) {
		rval = emmc_read(UPDFLAG_SECTOR_OFFSET + off, 1, (char *)upd_flag, 512);
		if (0 == rval)
			break;
		printk("emmc_read updflag sector(%lu) failed, try next\n", off);
	}

	data = (char *)upd_flag;
	for (j = 0; j < len; j++) {
		seq_printf(seq, "%c", *data);
		data++;
	}
	vfree(upd_flag);

	return 0;
}

static ssize_t provision_proc_store_updflag(struct file *file,
	const char __user *buffer, size_t count, loff_t *pos)
{
	int rval = 0;
	upd_flag_t *upd_flag;
	sector_t off;

	if (count > sizeof(upd_flag_t))
		return (-EINVAL);

	upd_flag = (upd_flag_t *)vmalloc(PAGE_SIZE);
	if (!upd_flag)
		return (-ENOMEM);

	memset(upd_flag, 0, sizeof(upd_flag_t));
	if (copy_from_user(upd_flag, buffer, count)) {
		vfree(upd_flag);
		return (-EFAULT);
	}

	for (off = 0; off < UPDFLAG_SECTOR_OFFSET; off++) {
		rval = emmc_write(UPDFLAG_SECTOR_OFFSET + off, 1, (char *)upd_flag, 512);
		if (0 == rval)
			break;
		printk("emmc_write updflag sector(%lu) failed, try next\n", off);
	}
	vfree(upd_flag);

	return count;
}

static int provision_proc_open_status(struct inode *inode, struct file *file)
{
	return single_open(file, provision_proc_show_status, NULL);
}

static const struct file_operations provision_proc_status_fops = {
	.open  = provision_proc_open_status,
	.read  = seq_read,
	.release = single_release,
	.owner = THIS_MODULE,
};

static int provision_proc_open_mmc(struct inode *inode, struct file *file)
{
	return single_open(file, provision_proc_show_mmc, PDE_DATA(inode));
}

static const struct file_operations provision_proc_mmc_fops = {
	.open  = provision_proc_open_mmc,
	.read  = seq_read,
	.write  = provision_proc_write_mmc,
	.release = single_release,
	.owner = THIS_MODULE,
};


static int provision_proc_open_updflag(struct inode *inode, struct file *file)
{
	return single_open(file, provision_proc_show_updflag, NULL);
}

static const struct file_operations provision_proc_updflag_fops = {
	.open  = provision_proc_open_updflag,
	.read  = seq_read,
	.write = provision_proc_store_updflag,
	.release = single_release,
	.owner = THIS_MODULE,
};

int provision_install_proc_entry_mmc(void)
{
	struct proc_dir_entry *provision_root_dir;
	struct proc_dir_entry   *ent;
	int i;
	/*unsigned int erase_size;
	erase_size = emmc_get_erase_size();*/

	for (i = 0; i < ARRAY_SIZE(provs); i++) {
		/*provs[i].start[0] = IMGHDR_START_SECTOR + (erase_size * 3) + i*(IMGHDR_SECTORS/ARRAY_SIZE(provs));*/
		provs[i].start = IMGHDR_SECTOR_OFFSET + i*(IMGHDR_SECTOR_SIZE/ARRAY_SIZE(provs));
	}

	provision_root_dir = proc_mkdir("provision", NULL);
	if (provision_root_dir != NULL) {
		struct proc_dir_entry	*dent;

		ent = proc_create("status", S_IFREG|S_IRUGO, provision_root_dir, &provision_proc_status_fops);
		if (!ent)
			printk("can not creat proc entry 'status' for provision!!\n");

		ent = proc_create("updflag", S_IFREG|S_IRUGO, provision_root_dir, &provision_proc_updflag_fops);
		if (!ent)
			printk("can not creat proc entry 'updflag' for provision!!\n");

		dent = proc_mkdir("partition", provision_root_dir);

		for (i = 0; i < ARRAY_SIZE(provs); i++)  {
			ent = proc_create_data(provs[i].name, S_IFREG|S_IRUGO|S_IWUSR, dent, &provision_proc_mmc_fops, &provs[i]);
			if (!ent) {
				continue;
			}
		}
	}

	return 0;
}

