/*
 * A V4L2 driver for NovaTek nt99141  cameras.
 *
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/videodev2.h>
#include <linux/clk.h>
#include <linux/gpio.h>
#include <media/v4l2-device.h>
#include <media/v4l2-chip-ident.h>
#include <media/v4l2-mediabus.h>//linux-3.0
#include <linux/io.h>
//#include <mach/gpio_v2.h>
#include <mach/sys_config.h>
#include <linux/regulator/consumer.h>
#include <mach/system.h>
#include "../../../../power/axp_power/axp-gpio.h"
#if defined CONFIG_ARCH_SUN4I
#include "../include/sun4i_csi_core.h"
#include "../include/sun4i_dev_csi.h"
#elif defined CONFIG_ARCH_SUN5I
#include "../include/sun5i_csi_core.h"
#include "../include/sun5i_dev_csi.h"
#elif defined CONFIG_ARCH_SUN7I
#include "../include/sunxi_csi_core.h"
#include "../include/sunxi_dev_csi.h"
#endif

MODULE_AUTHOR("lfzhu");
MODULE_DESCRIPTION("A low-level driver for nt99141 sensors");
MODULE_LICENSE("GPL");

//mount -o remount,rw /system
//./build.sh -p sun7i_android

//for internel driver debug
#define DEV_DBG_EN   		1 
#if(DEV_DBG_EN == 1)		
#define csi_dev_dbg(x,arg...) printk(KERN_INFO"[CSI_DEBUG][NT99141]"x,##arg)
#else
#define csi_dev_dbg(x,arg...) 
#endif
#define csi_dev_err(x,arg...) printk(KERN_INFO"[CSI_ERR][NT99141]"x,##arg)
#define csi_dev_print(x,arg...) printk(KERN_INFO"[CSI][NT99141]"x,##arg)

#define MCLK (24*1000*1000)
#define VREF_POL	CSI_HIGH
#define HREF_POL	CSI_HIGH
#define CLK_POL		CSI_RISING
#define IO_CFG		0						//0 for csi0

#define DRIVER_NAME  "nt99141"

//define the voltage level of control signal
#define CSI_STBY_ON			1
#define CSI_STBY_OFF 		0
#define CSI_RST_ON			0
#define CSI_RST_OFF			1
#define CSI_PWR_ON			1
#define CSI_PWR_OFF		0


#define V4L2_IDENT_SENSOR 0x1410

#define REG_ADDR_STEP 2
#define REG_DATA_STEP 1
#define REG_STEP 			(REG_ADDR_STEP+REG_DATA_STEP)


/*
 * Basic window sizes.  These probably belong somewhere more globally
 * useful.
 */
 
#define HD720_WIDTH 	1280
#define HD720_HEIGHT	720

#define HD720_960_WIDTH 	960
#define HD720_880_WIDTH 	880

#define WVGA_WIDTH       800
#define WVGA_HEIGHT      600

#define CIF4_WIDTH		704
#define CIF4_HEIGHT		576

#define VGA_WIDTH		640
#define VGA_HEIGHT	       480

#define CIF_WIDTH		352
#define CIF_HEIGHT		288

#define QVGA_WIDTH	       320
#define QVGA_HEIGHT	240

#define QCIF_WIDTH		176
#define QCIF_HEIGHT		144

/*
 * Our nominal (default) frame rate.
 */
#define SENSOR_FRAME_RATE 30

/*
 * The nt99141 sits on i2c with ID 0x54
 */
#define I2C_ADDR 0x54

/* Registers */

/*
 * Information we maintain about a known sensor.
 */
struct sensor_format_struct;  /* coming later */
__csi_subdev_info_t ccm_info_con = 
{
	.mclk 	= MCLK,
	.vref 	= VREF_POL,
	.href 	= HREF_POL,
	.clock	= CLK_POL,
	.iocfg	= IO_CFG,
};

struct sensor_info {
	struct v4l2_subdev sd;
	struct sensor_format_struct *fmt;  /* Current format */
	__csi_subdev_info_t *ccm_info;
	int	width;
	int	height;
	int brightness;
	int	contrast;
	int saturation;
	int hue;
	int hflip;
	int vflip;
	int gain;
	int autogain;
	int exp;
	enum v4l2_exposure_auto_type autoexp;
	int autowb;
	enum v4l2_whiteblance wb;
	enum v4l2_colorfx clrfx;
	enum v4l2_flash_mode flash_mode;
	enum v4l2_aetarget_ctrl aetarget;
	u8 clkrc;			/* Clock divider value */
};

static inline struct sensor_info *to_state(struct v4l2_subdev *sd)
{
	return container_of(sd, struct sensor_info, sd);
}


struct regval_list {
	unsigned char reg_num[REG_ADDR_STEP];
	unsigned char value[REG_DATA_STEP];
};
  
/*
 * The default register settings
 *
 */
static struct regval_list sensor_ae_target1_default_regs[] = 
{
	{{0x31,0x09},{0x04}},
	{{0x30,0x40},{0x04}},
	{{0x30,0x41},{0x02}},
	{{0x30,0x42},{0xFF}},
	{{0x30,0x43},{0x08}},
	{{0x30,0x52},{0xE0}},//F0
 // {{0x30,0x54},{0x00}},
 // {{0x30,0x55},{0x40}},
	{{0x30,0x5F},{0x33}},
	{{0x30,0x69},{0x01}},
	{{0x30,0x6A},{0x02}}, 
	{{0x31,0x00},{0x07}},
	{{0x31,0x05},{0x01}},
	{{0x31,0x06},{0x03}},
	{{0x31,0x08},{0x05}},
	{{0x31,0x10},{0x22}},
	{{0x31,0x11},{0x57}},
	{{0x31,0x12},{0x22}},
	{{0x31,0x13},{0x55}},
	{{0x31,0x14},{0x05}},
	{{0x31,0x35},{0x00}},
	{{0x32,0xF0},{0x01}},
	{{0x32,0x90},{0x01}},
	{{0x32,0x91},{0xA0}},
	{{0x32,0x96},{0x01}},
	{{0x32,0x97},{0x73}},

#if 0
    {{0x32,0x10},{0x1C}}, //0x16
    {{0x32,0x11},{0x1C}}, //0x19
    {{0x32,0x12},{0x1D}},//0x16
    {{0x32,0x13},{0x1A}},//0x14
    {{0x32,0x14},{0x12}},//0x15
    {{0x32,0x15},{0x13}},//0x18
    {{0x32,0x16},{0x13}},//0x15
    {{0x32,0x17},{0x12}},//0x14
    {{0x32,0x18},{0x12}},//0x15
    {{0x32,0x19},{0x13}},//0x18
    {{0x32,0x1A},{0x13}},//0x15
    {{0x32,0x1B},{0x12}},//0x14
    {{0x32,0x1C},{0x13}},//0x14
    {{0x32,0x1D},{0x14}},//0x17
    {{0x32,0x1E},{0x14}},//0x14
    {{0x32,0x1F},{0x14}},//0x12
#else
    {{0x32,0x10},{0x18}}, //0x16
    {{0x32,0x11},{0x18}}, //0x19
    {{0x32,0x12},{0x18}},//0x16
    {{0x32,0x13},{0x1A}},//0x14
    {{0x32,0x14},{0x10}},//0x15
    {{0x32,0x15},{0x11}},//0x18
    {{0x32,0x16},{0x13}},//0x15
    {{0x32,0x17},{0x12}},//0x14
    {{0x32,0x18},{0x10}},//0x15
    {{0x32,0x19},{0x11}},//0x18
    {{0x32,0x1A},{0x13}},//0x15
    {{0x32,0x1B},{0x12}},//0x14
    {{0x32,0x1C},{0x11}},//0x14
    {{0x32,0x1D},{0x10}},//0x17
    {{0x32,0x1E},{0x13}},//0x14
    {{0x32,0x1F},{0x13}},//0x12
#endif
	{{0x32,0x31},{0x74}},
	{{0x32,0x32},{0xC4}},
	{{0x33,0x26},{0x02}},
	{{0x33,0x27},{0x0A}},
	{{0x33,0x28},{0x0A}},
	{{0x33,0x29},{0x06}},
	{{0x33,0x2A},{0x06}},
	{{0x33,0x2B},{0x1C}},
	{{0x33,0x2C},{0x1C}},
	{{0x33,0x2D},{0x00}},
	{{0x33,0x2E},{0x1D}},
	{{0x33,0x2F},{0x1F}},
	{{0x32,0xF6},{0x0F}},
	{{0x32,0xF9},{0x84}}, //42
	{{0x32,0xFA},{0x48}}, //24
	{{0x33,0x25},{0x4A}},
	{{0x33,0x30},{0x00}},
	{{0x33,0x31},{0x0A}},
	{{0x33,0x32},{0xFF}},
	{{0x33,0x38},{0x30}},
	{{0x33,0x39},{0x84}},
	{{0x33,0x3A},{0x48}},
	{{0x33,0x3F},{0x07}},
	{{0x33,0x60},{0x0D}},
	{{0x33,0x61},{0x1C}},
	{{0x33,0x62},{0x31}},
	{{0x33,0x63},{0x37}},
	{{0x33,0x64},{0x8A}},
	{{0x33,0x65},{0x80}},//80
	{{0x33,0x66},{0x70}},//68
	{{0x33,0x67},{0x58}},//60
	{{0x33,0x68},{0x30}},
	{{0x33,0x69},{0x28}},
	{{0x33,0x6A},{0x20}},
	{{0x33,0x6B},{0x10}},
	{{0x33,0x6C},{0x00}},
	{{0x33,0x6D},{0x20}},
	{{0x33,0x6E},{0x1C}},
	{{0x33,0x6F},{0x18}},
	{{0x33,0x70},{0x10}},
	{{0x33,0x71},{0x38}},
	{{0x33,0x72},{0x3C}},
	{{0x33,0x73},{0x3F}},
	{{0x33,0x74},{0x3F}},
	{{0x33,0x8A},{0x34}},
	{{0x33,0x8B},{0x7F}},
	{{0x33,0x8C},{0x10}},
	{{0x33,0x8D},{0x23}},
	{{0x33,0x8E},{0x7F}},
	{{0x33,0x8F},{0x14}},
	{{0x33,0x75},{0x0A}},
	{{0x33,0x76},{0x0C}},
	{{0x33,0x77},{0x10}},
	{{0x33,0x78},{0x14}},
	
 #if 0 //For 3309A 
 	  //New gamma and CC
    {{0x32,0x70},{0x00}},
    {{0x32,0x71},{0x0c}},
    {{0x32,0x72},{0x18}},
    {{0x32,0x73},{0x2d}},
    {{0x32,0x74},{0x41}},
    {{0x32,0x75},{0x52}},
    {{0x32,0x76},{0x6d}},
    {{0x32,0x77},{0x82}},
    {{0x32,0x78},{0x93}},
    {{0x32,0x79},{0xa3}},
    {{0x32,0x7A},{0xbc}},
    {{0x32,0x7B},{0xd3}},
    {{0x32,0x7C},{0xE6}},
    {{0x32,0x7D},{0xF5}},
    {{0x32,0x7E},{0xFF}},
    
    {{0x33,0x02},{0x00}},
    {{0x33,0x03},{0x5c}},
    {{0x33,0x04},{0x00}},
    {{0x33,0x05},{0x6c}},
    {{0x33,0x06},{0x00}},
    {{0x33,0x07},{0x39}},
    {{0x33,0x08},{0x07}},
    {{0x33,0x09},{0xbe}},
    {{0x33,0x0A},{0x06}},
    {{0x33,0x0B},{0xf9}},
    {{0x33,0x0C},{0x01}},
    {{0x33,0x0D},{0x46}},
    {{0x33,0x0E},{0x01}},
    {{0x33,0x0F},{0x0d}},
    {{0x33,0x10},{0x06}},
    {{0x33,0x11},{0xfd}},
    {{0x33,0x12},{0x07}},
    {{0x33,0x13},{0xfc}},		 
    
   	{{0x32,0x50},{0xC0}},
  	{{0x32,0x51},{0x00}},
  	{{0x32,0x52},{0xDF}},
  	{{0x32,0x53},{0x48}},//0E
  	{{0x32,0x54},{0x00}},
  	{{0x32,0x55},{0xEB}},
  	{{0x32,0x56},{0x6C}},//81
  	{{0x32,0x58},{0x0C}},
  	{{0x32,0x9B},{0x31}},
  	{{0x32,0xA1},{0x00}},
  	{{0x32,0xA2},{0xA0}},
  	{{0x32,0xA3},{0x01}},
  	{{0x32,0xA4},{0xA0}},
  	{{0x32,0xA5},{0x01}},
  	{{0x32,0xA6},{0x18}},
  	{{0x32,0xA7},{0x01}},
  	{{0x32,0xA8},{0xE0}},
  #else //For 3208B
  
  {{0x32,0x70},{0x04}},//YW_TUNE
	{{0x32,0x71},{0x13}},
	{{0x32,0x72},{0x21}},
	{{0x32,0x73},{0x35}},
	{{0x32,0x74},{0x49}},
	{{0x32,0x75},{0x5A}},
	{{0x32,0x76},{0x76}},
	{{0x32,0x77},{0x88}},
	{{0x32,0x78},{0x9E}},
	{{0x32,0x79},{0xB0}},
	{{0x32,0x7A},{0xCA}},
	{{0x32,0x7B},{0xE2}},
	{{0x32,0x7C},{0xEF}},
	{{0x32,0x7D},{0xF9}},
	{{0x32,0x7E},{0xFF}},
  
  {{0x33,0x02},{0x00}}, //CC CAL 160Tool
  {{0x33,0x03},{0x4C}},
  {{0x33,0x04},{0x00}},
  {{0x33,0x05},{0x96}},
  {{0x33,0x06},{0x00}},
  {{0x33,0x07},{0x1D}},
  {{0x33,0x08},{0x07}},
  {{0x33,0x09},{0xBB}},
  {{0x33,0x0A},{0x06}},
  {{0x33,0x0B},{0xDC}},
  {{0x33,0x0C},{0x01}},
  {{0x33,0x0D},{0x6A}},
  {{0x33,0x0E},{0x01}},
  {{0x33,0x0F},{0x10}},
  {{0x33,0x10},{0x07}},
  {{0x33,0x11},{0x20}},
  {{0x33,0x12},{0x07}},
  {{0x33,0x13},{0xD0}},
	
	{{0x32,0x50},{0xC8}}, 
  {{0x32,0x51},{0x01}},
  {{0x32,0x52},{0x1B}},
  {{0x32,0x53},{0x85}},
  {{0x32,0x54},{0x00}},
  {{0x32,0x55},{0xC9}},
  {{0x32,0x56},{0x91}},
  {{0x32,0x57},{0x48}},
  {{0x32,0x58},{0x0A}},
  {{0x32,0x9B},{0x31}},
  {{0x32,0xA1},{0x00}},
  {{0x32,0xA2},{0xFF}},
  {{0x32,0xA3},{0x01}},
  {{0x32,0xA4},{0xB0}},
  {{0x32,0xA5},{0x01}},
  {{0x32,0xA6},{0x50}},
  {{0x32,0xA7},{0x01}},
  {{0x32,0xA8},{0xE0}},

  {{0x32,0xA9},{0x11}}, // +R  -B
  {{0x32,0xAA},{0x01}},
 #endif    
    {{0x30,0x53},{0x10}},
    {{0x30,0x54},{0x05}},
    {{0x32,0xF2},{0x7A}},
    {{0x32,0xFC},{0x0A}},       
 

	//CY Hsu 20131014 Start
	{{0x32,0x01},{0x7F}},
	{{0x32,0xB1},{0x69}}, //7D
	{{0x32,0xB2},{0x69}}, //7D
	{{0x32,0x57},{0x58}}, //60
	
#if 1
	
  {{0x32,0xB8}, {0x3F}},  // AE Target 0x38
  {{0x32,0xB9}, {0x31}}, 
  {{0x32,0xBB}, {0x87}}, 
  {{0x32,0xBC}, {0x38}}, 
  {{0x32,0xBD}, {0x3C}}, 
  {{0x32,0xBE}, {0x34}}, 
#elif 0
  
  {{0x32,0xB8}, {0x3B}},  // AE Target 0x34
  {{0x32,0xB9}, {0x2D}}, 
  {{0x32,0xBB}, {0x87}}, 
  {{0x32,0xBC}, {0x34}}, 
  {{0x32,0xBD}, {0x38}}, 
  {{0x32,0xBE}, {0x30}}, 
#else
  
  {{0x32,0xB8}, {0x36}}, // AE Target 0x30
  {{0x32,0xB9}, {0x2A}}, 
  {{0x32,0xBB}, {0x87}}, 
  {{0x32,0xBC}, {0x30}}, 
  {{0x32,0xBD}, {0x33}}, 
  {{0x32,0xBE}, {0x2D}}, 
#endif

	
	{{0x30,0x60},{0x01}},  
	//CY Hsu 20131014 End
};


//1280x720, 30fps,pclk=74Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_hd720_30fps_regs[] = 
{
  {{0x32,0xBF}, {0x60}}, 
  {{0x32,0xC0}, {0x5A}}, 
  {{0x32,0xC1}, {0x5A}}, 
  {{0x32,0xC2}, {0x5A}}, 
  {{0x32,0xC3}, {0x00}}, 
  {{0x32,0xC4}, {0x34}}, 
  {{0x32,0xC5}, {0x28}}, 
  {{0x32,0xC6}, {0x28}}, 
  {{0x32,0xC7}, {0x00}}, 
  {{0x32,0xC8}, {0xDE}}, 
  {{0x32,0xC9}, {0x5A}}, 
  {{0x32,0xCA}, {0x82}}, 
  {{0x32,0xCB}, {0x82}}, 
  {{0x32,0xCC}, {0x82}}, 
  {{0x32,0xCD}, {0x82}}, 
  {{0x32,0xDB}, {0x7B}}, 
  {{0x32,0xE0}, {0x05}}, 
  {{0x32,0xE1}, {0x00}}, 
  {{0x32,0xE2}, {0x02}}, 
  {{0x32,0xE3}, {0xD0}}, 
  {{0x32,0xE4}, {0x00}}, 
  {{0x32,0xE5}, {0x00}}, 
  {{0x32,0xE6}, {0x00}}, 
  {{0x32,0xE7}, {0x00}}, 
  {{0x32,0x00}, {0x3E}}, 
  //{{0x32,0x01}, {0x0F}}, 
  {{0x30,0x28}, {0x13}}, 
  {{0x30,0x29}, {0x20}}, 
  {{0x30,0x2A}, {0x00}}, 
  {{0x30,0x22}, {0x24}}, 
  {{0x30,0x23}, {0x24}}, 
  {{0x30,0x02}, {0x00}}, 
  {{0x30,0x03}, {0x04}}, 
  {{0x30,0x04}, {0x00}}, 
  {{0x30,0x05}, {0x04}}, 
  {{0x30,0x06}, {0x05}}, 
  {{0x30,0x07}, {0x03}}, 
  {{0x30,0x08}, {0x02}}, 
  {{0x30,0x09}, {0xD3}}, 
  {{0x30,0x0A}, {0x07}}, 
  {{0x30,0x0B}, {0x08}}, 
  {{0x30,0x0C}, {0x02}}, 
  {{0x30,0x0D}, {0xE4}}, 
  {{0x30,0x0E}, {0x05}}, 
  {{0x30,0x0F}, {0x00}}, 
  {{0x30,0x10}, {0x02}}, 
  {{0x30,0x11}, {0xD0}}, 
  //{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
 //{{0x32,0xBB}, {0x87}}, 
 //{{0x32,0xBC}, {0x38}}, 
 //{{0x32,0xBD}, {0x3C}}, 
 //{{0x32,0xBE}, {0x34}}, 
  {{0x32,0x01}, {0x7F}}, 
  //{{0x30,0x21}, {0x06}},   
  {{0x32,0x0A}, {0xB2}},
  {{0x30,0x60}, {0x01}}, 
};

//1280x720, 25fps,pclk=70Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_hd720_25fps_regs[] = 
{
    {{0x32,0xBF}, {0x60}},   
    {{0x32,0xC0}, {0x60}},   
    {{0x32,0xC1}, {0x60}},   
    {{0x32,0xC2}, {0x60}},   
    {{0x32,0xC3}, {0x00}},   
    {{0x32,0xC4}, {0x34}},   
    {{0x32,0xC5}, {0x28}},   
    {{0x32,0xC6}, {0x28}},   
    {{0x32,0xC7}, {0x00}},   
    {{0x32,0xC8}, {0xB8}},   
    {{0x32,0xC9}, {0x60}},   
    {{0x32,0xCA}, {0x88}},   
    {{0x32,0xCB}, {0x88}},   
    {{0x32,0xCC}, {0x88}},   
    {{0x32,0xCD}, {0x88}},   
    {{0x32,0xDB}, {0x77}},   
    {{0x32,0xE0}, {0x05}},   
    {{0x32,0xE1}, {0x00}},   
    {{0x32,0xE2}, {0x02}},   
    {{0x32,0xE3}, {0xD0}},   
    {{0x32,0xE4}, {0x00}},   
    {{0x32,0xE5}, {0x00}},   
    {{0x32,0xE6}, {0x00}},   
    {{0x32,0xE7}, {0x00}},   
    {{0x32,0x00}, {0x3E}},   
    //{{0x32,0x01}, {0x0F}}, 
    {{0x30,0x28}, {0x22}},   
    {{0x30,0x29}, {0x20}},   
    {{0x30,0x2A}, {0x04}},   
    {{0x30,0x22}, {0x24}},   
    {{0x30,0x23}, {0x24}},   
    {{0x30,0x02}, {0x00}},   
    {{0x30,0x03}, {0x04}},   
    {{0x30,0x04}, {0x00}},   
    {{0x30,0x05}, {0x04}},   
    {{0x30,0x06}, {0x05}},   
    {{0x30,0x07}, {0x03}},   
    {{0x30,0x08}, {0x02}},   
    {{0x30,0x09}, {0xD3}},   
    {{0x30,0x0A}, {0x07}},   
    {{0x30,0x0B}, {0x6C}},   
    {{0x30,0x0C}, {0x02}},   
    {{0x30,0x0D}, {0xE0}},   
    {{0x30,0x0E}, {0x05}},   
    {{0x30,0x0F}, {0x00}},   
    {{0x30,0x10}, {0x02}},   
    {{0x30,0x11}, {0xD0}},   
    //{{0x32,0xB8}, {0x3F}},   
  //{{0x32,0xB9}, {0x31}},   
    //{{0x32,0xBB}, {0x87}},   
   //{{0x32,0xBC}, {0x38}},   
   //{{0x32,0xBD}, {0x3C}},   
   //{{0x32,0xBE}, {0x34}},   
    //{{0x32,0x01}, {0x7F}}, 
    {{0x30,0x21}, {0x06}},   
    {{0x32,0x0A}, {0xB2}},   
    {{0x30,0x60}, {0x01}},   

};

//1280x720, 15fps,pclk=40Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_hd720_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x78}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x6D}}, 
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x27}}, 
{{0x30,0x29}, {0x20}}, 
{{0x30,0x2A}, {0x08}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0x04}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x05}}, 
{{0x30,0x07}, {0x03}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x06}}, 
{{0x30,0x0B}, {0x83}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x1F}}, 
{{0x30,0x0E}, {0x05}}, 
{{0x30,0x0F}, {0x00}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//1280x720, 5fps,pclk=16Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_hd720_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x84}}, 
{{0x32,0xC1}, {0x84}}, 
{{0x32,0xC2}, {0x84}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x2A}}, 
{{0x32,0xC9}, {0x84}}, 
{{0x32,0xCA}, {0xAC}}, 
{{0x32,0xCB}, {0xAC}}, 
{{0x32,0xCC}, {0xAC}}, 
{{0x32,0xCD}, {0xAC}}, 
{{0x32,0xDB}, {0x55}}, 
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
////CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}},
//{{0x32,0x01}, {0x0F}},
{{0x30,0x28}, {0x07}},
{{0x30,0x29}, {0x20}},
{{0x30,0x2A}, {0x04}},
{{0x30,0x22}, {0x24}},
{{0x30,0x23}, {0x24}},
{{0x30,0x02}, {0x00}},
{{0x30,0x03}, {0x04}},
{{0x30,0x04}, {0x00}},
{{0x30,0x05}, {0x04}},
{{0x30,0x06}, {0x05}},
{{0x30,0x07}, {0x03}},
{{0x30,0x08}, {0x02}},
{{0x30,0x09}, {0xD3}},
{{0x30,0x0A}, {0x07}},
{{0x30,0x0B}, {0x6C}},
{{0x30,0x0C}, {0x03}},
{{0x30,0x0D}, {0x4A}},
{{0x30,0x0E}, {0x05}},
{{0x30,0x0F}, {0x00}},
{{0x30,0x10}, {0x02}},
{{0x30,0x11}, {0xD0}},
//{{0x32,0xB8}, {0x3F}},
//{{0x32,0xB9}, {0x31}},
//{{0x32,0xBB}, {0x17}},
//{{0x32,0xBC}, {0x38}},
//{{0x32,0xBD}, {0x3C}},
//{{0x32,0xBE}, {0x34}},
//{{0x32,0x01}, {0x3F}},
{{0x30,0x21}, {0x06}},
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};


//[70Mhz YUYV_1280x720_15_25Fps]
static struct regval_list sensor_ae_target1_hd720_29fps_regs[] = 
{
    {{0x32,0xBF}, {0x60}},   
    {{0x32,0xC0}, {0x6A}},   
    {{0x32,0xC1}, {0x6A}},   
    {{0x32,0xC2}, {0x6A}},   
    {{0x32,0xC3}, {0x00}},   
    {{0x32,0xC4}, {0x34}},   
    {{0x32,0xC5}, {0x28}},   
    {{0x32,0xC6}, {0x28}},   
    {{0x32,0xC7}, {0x00}},   
    {{0x32,0xC8}, {0xB8}},   
    {{0x32,0xC9}, {0x6A}},   
    {{0x32,0xCA}, {0x92}},   
    {{0x32,0xCB}, {0x92}},   
    {{0x32,0xCC}, {0x92}},   
    {{0x32,0xCD}, {0x92}},   
    {{0x32,0xDB}, {0x77}},   
    {{0x32,0xE0}, {0x05}},   
    {{0x32,0xE1}, {0x00}},   
    {{0x32,0xE2}, {0x02}},   
    {{0x32,0xE3}, {0xD0}},   
    {{0x32,0xE4}, {0x00}},   
    {{0x32,0xE5}, {0x00}},   
    {{0x32,0xE6}, {0x00}},   
    {{0x32,0xE7}, {0x00}},   
    {{0x32,0x00}, {0x3E}},   
    //{{0x32,0x01}, {0x0F}}, 
    {{0x30,0x28}, {0x22}},   
    {{0x30,0x29}, {0x20}},   
    {{0x30,0x2A}, {0x04}},   
    {{0x30,0x22}, {0x24}},   
    {{0x30,0x23}, {0x24}},   
    {{0x30,0x02}, {0x00}},   
    {{0x30,0x03}, {0x04}},   
    {{0x30,0x04}, {0x00}},   
    {{0x30,0x05}, {0x04}},   
    {{0x30,0x06}, {0x05}},   
    {{0x30,0x07}, {0x03}},   
    {{0x30,0x08}, {0x02}},   
    {{0x30,0x09}, {0xD3}},   
    {{0x30,0x0A}, {0x07}},   
    {{0x30,0x0B}, {0x6C}},   
    {{0x30,0x0C}, {0x02}},   
    {{0x30,0x0D}, {0xE0}},   
    {{0x30,0x0E}, {0x05}},   
    {{0x30,0x0F}, {0x00}},   
    {{0x30,0x10}, {0x02}},   
    {{0x30,0x11}, {0xD0}},   
    //{{0x32,0xB8}, {0x3F}},   
  //{{0x32,0xB9}, {0x31}},   
    //{{0x32,0xBB}, {0x87}},   
   //{{0x32,0xBC}, {0x38}},   
   //{{0x32,0xBD}, {0x3C}},   
   //{{0x32,0xBE}, {0x34}},   
    //{{0x32,0x01}, {0x7F}}, 
    {{0x30,0x21}, {0x06}},   
    {{0x32,0x0A}, {0xB2}},   
    {{0x30,0x60}, {0x01}},
};

//[1280x720_30fps_50Hz]
static struct regval_list sensor_ae_target1_hd720_30fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
};

//[1280x720_30fps_60Hz]
static struct regval_list sensor_ae_target1_hd720_30fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x5F}}, 
{{0x32,0xC2}, {0x5F}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB9}}, 
{{0x32,0xC9}, {0x5F}}, 
{{0x32,0xCA}, {0x87}}, 
{{0x32,0xCB}, {0x87}}, 
{{0x32,0xCC}, {0x87}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x77}}, 
};

///1280,720 25FPS [50Hz]
static struct regval_list sensor_ae_target1_hd720_25fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x60}}, 
{{0x32,0xC2}, {0x60}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x60}}, 
{{0x32,0xCA}, {0x88}}, 
{{0x32,0xCB}, {0x88}}, 
{{0x32,0xCC}, {0x88}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x77}}, 

};

static struct regval_list sensor_ae_target1_hd720_25fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x63}}, 
{{0x32,0xC1}, {0x63}}, 
{{0x32,0xC2}, {0x63}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x9A}}, 
{{0x32,0xC9}, {0x63}}, 
{{0x32,0xCA}, {0x8B}}, 
{{0x32,0xCB}, {0x8B}}, 
{{0x32,0xCC}, {0x8B}}, 
{{0x32,0xCD}, {0x8B}}, 
{{0x32,0xDB}, {0x73}}, 

};

//1280,720 15FPS [50Hz]
static struct regval_list sensor_ae_target1_hd720_15fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x78}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x6D}}, 
};

static struct regval_list sensor_ae_target1_hd720_15fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x64}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x68}}, 
};

//1280,720 5FPS [50Hz]
static struct regval_list sensor_ae_target1_hd720_5fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x84}}, 
{{0x32,0xC1}, {0x84}}, 
{{0x32,0xC2}, {0x84}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x2A}}, 
{{0x32,0xC9}, {0x84}}, 
{{0x32,0xCA}, {0xAC}}, 
{{0x32,0xCB}, {0xAC}}, 
{{0x32,0xCC}, {0xAC}}, 
{{0x32,0xCD}, {0xAC}}, 
{{0x32,0xDB}, {0x55}}, 
};

static struct regval_list sensor_ae_target1_hd720_5fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x88}}, 
{{0x32,0xC1}, {0x88}}, 
{{0x32,0xC2}, {0x88}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x23}}, 
{{0x32,0xC9}, {0x88}}, 
{{0x32,0xCA}, {0xB0}}, 
{{0x32,0xCB}, {0xB0}}, 
{{0x32,0xCC}, {0xB0}}, 
{{0x32,0xCD}, {0xB0}}, 
{{0x32,0xDB}, {0x51}}, 
};

//[1280x720_50Hz]
static struct regval_list sensor_ae_target1_hd720_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}},
{{0x32,0xC0}, {0x6A}},
{{0x32,0xC1}, {0x6A}},
{{0x32,0xC2}, {0x6A}},
{{0x32,0xC3}, {0x00}},
{{0x32,0xC4}, {0x34}},
{{0x32,0xC5}, {0x28}},
{{0x32,0xC6}, {0x28}},
{{0x32,0xC7}, {0x00}},
{{0x32,0xC8}, {0xB8}},
{{0x32,0xC9}, {0x6A}},
{{0x32,0xCA}, {0x92}},
{{0x32,0xCB}, {0x92}},
{{0x32,0xCC}, {0x92}},
{{0x32,0xCD}, {0x92}},
{{0x32,0xDB}, {0x77}},
};

//[1280x720_60Hz]
static struct regval_list sensor_ae_target1_hd720_29fps_60hz_regs[] =  
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x9A}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};

//YUYV_960x720_30.00Fps
static struct regval_list sensor_ae_target1_960hd720_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_960x720_25.00Fps]
static struct regval_list sensor_ae_target1_960hd720_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}}, 
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_960x720_15.00Fps]
static struct regval_list sensor_ae_target1_960hd720_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_960x720_5.00Fps]
static struct regval_list sensor_ae_target1_960hd720_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}},  
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};


//[YUYV_960x720_15_25Fps]
static struct regval_list sensor_ae_target1_960hd720_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}},
};

//[960x720_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_960hd720_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}},
};

//[960x720_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_960hd720_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//[YUYV_880x720_30.00Fps]
static struct regval_list sensor_ae_target1_880hd720_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_880x720_25.00Fps
static struct regval_list sensor_ae_target1_880hd720_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_880x720_15.00Fps
static struct regval_list sensor_ae_target1_880hd720_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}}, 
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_880x720_5.00Fps
static struct regval_list sensor_ae_target1_880hd720_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}},  
//CY Hsu 20131014 End
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_880x720_15_25_Fps]
static struct regval_list sensor_ae_target1_880hd720_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x07}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x18}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}},
};

//[880x720_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_880hd720_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

static struct regval_list sensor_ae_target1_880hd720_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};


//[YUYV_800x600_15_25_Fps]
static struct regval_list sensor_ae_target1_wvga_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x20}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x58}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x33}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x33}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x48}},
{{0x30,0x60}, {0x01}}, 
};

//[800x600_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_wvga_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[800x600_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_wvga_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//[YUYV_704x576_30.00Fps]
static struct regval_list sensor_ae_target1_4cif_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV, 704x576_25.00Fps
static struct regval_list sensor_ae_target1_4cif_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV, 704x576_15.00Fps
static struct regval_list sensor_ae_target1_4cif_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV, 704x576_5.00Fps
static struct regval_list sensor_ae_target1_4cif_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};
      
    
//[YUYV_704x576_25_30_Fps]
static struct regval_list sensor_ae_target1_4cif_29fps_regs[] =
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x07}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x18}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x58}},
{{0x30,0x60}, {0x01}}, 
};

//[704x576_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_4cif_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[704x576_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_4cif_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//640x480, 30fps,pclk=80Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_vga_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//640x480, 25fps,pclk=80Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_vga_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//640x480, 15fps,pclk=40Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_vga_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//640x480, 5fps,pclk=16Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target1_vga_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_640x480_25_30_Fps]
static struct regval_list sensor_ae_target1_vga_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}},
};

//[640x480_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_vga_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[640x480_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_vga_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//[YUYV_352x288_30.00Fps]
static struct regval_list sensor_ae_target1_cif_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_352x288_25.00Fps]
static struct regval_list sensor_ae_target1_cif_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_352x288_15.00Fps]
static struct regval_list sensor_ae_target1_cif_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_352x288_5.00Fps
static struct regval_list sensor_ae_target1_cif_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};
      
//[YUYV_352x288_15_25_Fps]
static struct regval_list sensor_ae_target1_cif_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}}, 
};

//[352x288_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_cif_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[352x288_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_cif_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};

//320x240
static struct regval_list sensor_ae_target1_qvga_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0},{0x01}},
{{0x32,0xE1},{0x40}},
{{0x32,0xE2},{0x00}},
{{0x32,0xE3},{0xF0}},
{{0x32,0xE4},{0x02}},
{{0x32,0xE5},{0x02}},
{{0x32,0xE6},{0x02}},
{{0x32,0xE7},{0x03}},
{{0x32,0x00},{0x3F}},
//{{0x32,0x01},{0x0F}},
{{0x30,0x28},{0x09}},
{{0x30,0x29},{0x00}},
{{0x30,0x2A},{0x04}},
{{0x30,0x22},{0x24}},
{{0x30,0x23},{0x24}},
{{0x30,0x02},{0x00}},
{{0x30,0x03},{0xA4}},
{{0x30,0x04},{0x00}},
{{0x30,0x05},{0x04}},
{{0x30,0x06},{0x04}},
{{0x30,0x07},{0x63}},
{{0x30,0x08},{0x02}},
{{0x30,0x09},{0xD3}},
{{0x30,0x0A},{0x05}},
{{0x30,0x0B},{0x47}},
{{0x30,0x0C},{0x02}},
{{0x30,0x0D},{0xE4}},
{{0x30,0x0E},{0x03}},
{{0x30,0x0F},{0xC0}},
{{0x30,0x10},{0x02}},
{{0x30,0x11},{0xD0}},
//{{0x32,0xB8},{0x3F}},
//{{0x32,0xB9},{0x31}},
//{{0x32,0xBB},{0x17}},
//{{0x32,0xBC},{0x38}},
//{{0x32,0xBD},{0x3C}},
//{{0x32,0xBE},{0x34}},
//{{0x32,0x01},{0x7F}},
{{0x30,0x21},{0x06}},
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60},{0x01}},
};

static struct regval_list sensor_ae_target1_qvga_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

static struct regval_list sensor_ae_target1_qvga_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//320x240
static struct regval_list sensor_ae_target1_qvga_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//[YUYV_320x240_15_25_Fps]
static struct regval_list sensor_ae_target1_qvga_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}}, 
};

//[320x240_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_qvga_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[320x240_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_qvga_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x90}}, 
{{0x32,0xCB}, {0x90}}, 
{{0x32,0xCC}, {0x90}}, 
{{0x32,0xCD}, {0x90}}, 
{{0x32,0xDB}, {0x73}}, 
};

//[YUYV_176x144_30Fps]
static struct regval_list sensor_ae_target1_qcif_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};


//[YUYV_176x144_25Fps]
static struct regval_list sensor_ae_target1_qcif_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_176x144_15.00]
static struct regval_list sensor_ae_target1_qcif_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_176x144_5.00]
static struct regval_list sensor_ae_target1_qcif_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x3F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_176x144_25_30_Fps]
static struct regval_list sensor_ae_target1_qcif_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x7C}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x3F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
//{{0x32,0xB8}, {0x3F}}, 
//{{0x32,0xB9}, {0x31}}, 
//{{0x32,0xBB}, {0x17}}, 
//{{0x32,0xBC}, {0x38}}, 
//{{0x32,0xBD}, {0x3C}}, 
//{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}}, 
};

//[176x144_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target1_qcif_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[176x144_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target1_qcif_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};

//960x720,880x720,800x600,704x576,640x480,352x288,320x240,176x144 Anti Flicker]
//[50Hz]
static struct regval_list sensor_ae_target1_default_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
};

//[60Hz]
static struct regval_list sensor_ae_target1_default_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x5F}}, 
{{0x32,0xC2}, {0x5F}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x34}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB9}}, 
{{0x32,0xC9}, {0x5F}}, 
{{0x32,0xCA}, {0x87}}, 
{{0x32,0xCB}, {0x87}}, 
{{0x32,0xCC}, {0x87}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x77}}, 
};

static struct regval_list sensor_oe_disable[] =
{
	
};

/*
 * The white balance settings
 * Here only tune the R G B channel gain. 
 * The white balance enalbe bit is modified in sensor_s_autowb and sensor_s_wb
 */
static struct regval_list sensor_wb_auto_regs[] = {
	{{0x32,0x01},{0x7F}},
};

static struct regval_list sensor_wb_cloud_regs[] = {
	{{0x32,0x90}, {0x01}},	// 337 / 256 = 1.316                                 
	{{0x32,0x91}, {0x51}},                                                     
	{{0x32,0x96}, {0x01}},	// 256 / 256 = 1.0                                   
	{{0x32,0x97}, {0x00}},                                                     
	{{0x30,0x60}, {0x01}},
};

//daylight 6300k
static struct regval_list sensor_wb_daylight_regs[] = {
	{{0x32,0x90}, {0x01}},	// 312 / 256 = 1.218                                 
	{{0x32,0x91}, {0x67}},                                                     
	{{0x32,0x96}, {0x01}},	// 360 / 256 = 1.406                                 
	{{0x32,0x97}, {0x3d}},                                                     
	{{0x30,0x60}, {0x01}}, 
};

//incandescence 3000k
static struct regval_list sensor_wb_incandescence_regs[] = {
	//bai re guang or bai chi deng
	{{0x32,0x90}, {0x00}},	// 304 / 256 = 1.187                                 
	{{0x32,0x91}, {0xe4}},                                                     
	{{0x32,0x96}, {0x01}},	// 459 / 256 = 1.793                                 
	{{0x32,0x97}, {0xe0}},                                                     
	{{0x30,0x60}, {0x01}},

};

//fluorescent 3800k
static struct regval_list sensor_wb_fluorescent_regs[] = {
	//ri guang deng or ying guang deng
	{{0x32,0x90}, {0x01}},	// 368 / 256 = 1.437                                 
	{{0x32,0x91}, {0x1d}},                                                     
	{{0x32,0x96}, {0x01}},	// 511 / 256 = 1.996                                 
	{{0x32,0x97}, {0xb4}},                                                     
	{{0x30,0x60}, {0x01}}, 
};

//tungsten 2600k
static struct regval_list sensor_wb_tungsten_regs[] = {
	//wu si deng
	{{0x32,0x90}, {0x00}},	// 256 / 256 = 1.0                                   
	{{0x32,0x91}, {0xe6}},                                                     
	{{0x32,0x96}, {0x02}},	// 560 / 256 = 2.1875                                
	{{0x32,0x97}, {0xe0}},                                                     
	{{0x30,0x60}, {0x01}},
};

/*
 * The color effect settings
 */
static struct regval_list sensor_colorfx_none_regs[] = {
	
};

static struct regval_list sensor_colorfx_bw_regs[] = {
	
};

static struct regval_list sensor_colorfx_sepia_regs[] = {
	
};

static struct regval_list sensor_colorfx_negative_regs[] = {
	
};

static struct regval_list sensor_colorfx_emboss_regs[] = {
//NULL	
};

static struct regval_list sensor_colorfx_sketch_regs[] = {
//NULL
};

static struct regval_list sensor_colorfx_sky_blue_regs[] = {
	
};

static struct regval_list sensor_colorfx_grass_green_regs[] = {
	
};

static struct regval_list sensor_colorfx_skin_whiten_regs[] = {
//NULL	
};

static struct regval_list sensor_colorfx_vivid_regs[] = {
//NULL
};

/*
 * The brightness setttings
 */
static struct regval_list sensor_brightness_neg4_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_neg3_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_neg2_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_neg1_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_zero_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_pos1_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_pos2_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_pos3_regs[] = {
//NULL
};

static struct regval_list sensor_brightness_pos4_regs[] = {
//NULL
};

/*
 * The contrast setttings
 */
static struct regval_list sensor_contrast_neg4_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_neg3_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_neg2_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_neg1_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_zero_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_pos1_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_pos2_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_pos3_regs[] = {
//NULL
};

static struct regval_list sensor_contrast_pos4_regs[] = {
//NULL
};

/*
 * The saturation setttings
 */
static struct regval_list sensor_saturation_neg4_regs[] = {
{{0x32,0xF3}, {0x50}},
};

static struct regval_list sensor_saturation_neg3_regs[] = {
{{0x32,0xF3}, {0x60}},
};

static struct regval_list sensor_saturation_neg2_regs[] = {
{{0x32,0xF3}, {0x70}},
};

static struct regval_list sensor_saturation_neg1_regs[] = {
{{0x32,0xF3}, {0x80}},
};

static struct regval_list sensor_saturation_zero_regs[] = {
{{0x32,0xF3}, {0x90}},
};

static struct regval_list sensor_saturation_pos1_regs[] = {
{{0x32,0xF3}, {0xA0}},
};

static struct regval_list sensor_saturation_pos2_regs[] = {
{{0x32,0xF3}, {0xB0}},
};

static struct regval_list sensor_saturation_pos3_regs[] = {
{{0x32,0xF3}, {0xC0}},
};

static struct regval_list sensor_saturation_pos4_regs[] = {
{{0x32,0xF3}, {0xE0}},
};

/*
 * The exposure target setttings
 */
static struct regval_list sensor_ev_neg4_regs[] = {
	 {{0x32,0xF2}, {0x3A}},              	
};

static struct regval_list sensor_ev_neg3_regs[] = {
	  {{0x32,0xF2}, {0x4A}},                 	
};

static struct regval_list sensor_ev_neg2_regs[] = {
	  {{0x32,0xF2}, {0x5A}},               	
};

static struct regval_list sensor_ev_neg1_regs[] = {
	  {{0x32,0xF2}, {0x6A}},             	
};

static struct regval_list sensor_ev_zero_regs[] = {
	  {{0x32,0xF2}, {0x7A}},              	
};

static struct regval_list sensor_ev_pos1_regs[] = {
	  {{0x32,0xF2}, {0x8A}},         	
};

static struct regval_list sensor_ev_pos2_regs[] = {
	 {{0x32,0xF2}, {0x9A}},             	
};

static struct regval_list sensor_ev_pos3_regs[] = {
    {{0x32,0xF2}, {0xAA}},          	
};

static struct regval_list sensor_ev_pos4_regs[] = {
	 {{0x32,0xF2}, {0xBA}},             
};

static struct regval_list sensor_ae_target0_default_regs[] = 
{
	{{0x31,0x09},{0x04}},
	{{0x30,0x40},{0x04}},
	{{0x30,0x41},{0x02}},
	{{0x30,0x42},{0xFF}},
	{{0x30,0x43},{0x08}},
	{{0x30,0x52},{0xE0}},
//  {{0x30,0x54},{0x00}},
//  {{0x30,0x55},{0x40}},
	{{0x30,0x5F},{0x33}},
	{{0x30,0x69},{0x01}},
	{{0x30,0x6A},{0x02}}, 
	{{0x31,0x00},{0x07}},
	{{0x31,0x05},{0x01}},
	{{0x31,0x06},{0x03}},
	{{0x31,0x08},{0x05}},
	{{0x31,0x10},{0x22}},
	{{0x31,0x11},{0x57}},
	{{0x31,0x12},{0x22}},
	{{0x31,0x13},{0x55}},
	{{0x31,0x14},{0x05}},
	{{0x31,0x35},{0x00}},
	{{0x32,0xF0},{0x01}},
	{{0x32,0x90},{0x01}},
	{{0x32,0x91},{0xA0}},
	{{0x32,0x96},{0x01}},
	{{0x32,0x97},{0x73}},
#if 1 // for 3208B
	{{0x32,0x50},{0xC8}}, 
  {{0x32,0x51},{0x01}},
  {{0x32,0x52},{0x1B}},
  {{0x32,0x53},{0x85}},
  {{0x32,0x54},{0x00}},
  {{0x32,0x55},{0xC9}},
  {{0x32,0x56},{0x91}},
  {{0x32,0x57},{0x48}},
  {{0x32,0x58},{0x0A}},
  {{0x32,0x9B},{0x31}},
  {{0x32,0xA1},{0x00}},
  {{0x32,0xA2},{0xFF}},
  {{0x32,0xA3},{0x01}},
  {{0x32,0xA4},{0xB0}},
  {{0x32,0xA5},{0x01}},
  {{0x32,0xA6},{0x50}},
  {{0x32,0xA7},{0x01}},
  {{0x32,0xA8},{0xE0}},

  {{0x32,0xA9},{0x11}}, // +R  -B
  {{0x32,0xAA},{0x01}}, 
#else // for 3309A
	{{0x32,0x50},{0xC0}},
	{{0x32,0x51},{0x00}},
	{{0x32,0x52},{0xDF}},
	{{0x32,0x53},{0x0E}},
	{{0x32,0x54},{0x00}},
	{{0x32,0x55},{0xEB}},
	{{0x32,0x56},{0x81}},
	{{0x32,0x57},{0x40}},	
	{{0x32,0x9B},{0x31}},
	{{0x32,0xA1},{0x00}},
	{{0x32,0xA2},{0xA0}},
	{{0x32,0xA3},{0x01}},
	{{0x32,0xA4},{0xA0}},
	{{0x32,0xA5},{0x01}},
	{{0x32,0xA6},{0x18}},
	{{0x32,0xA7},{0x01}},
	{{0x32,0xA8},{0xE0}},
#endif
	{{0x32,0x10},{0x16}},
	{{0x32,0x11},{0x19}},
	{{0x32,0x12},{0x16}},
	{{0x32,0x13},{0x14}},
	{{0x32,0x14},{0x15}},
	{{0x32,0x15},{0x18}},
	{{0x32,0x16},{0x15}},
	{{0x32,0x17},{0x14}},
	{{0x32,0x18},{0x15}},
	{{0x32,0x19},{0x18}},
	{{0x32,0x1A},{0x15}},
	{{0x32,0x1B},{0x14}},
	{{0x32,0x1C},{0x14}},
	{{0x32,0x1D},{0x17}},
	{{0x32,0x1E},{0x14}},
	{{0x32,0x1F},{0x12}},
	{{0x32,0x31},{0x74}},
	{{0x32,0x32},{0xC4}},
	{{0x33,0x26},{0x02}},
	{{0x33,0x27},{0x0A}},
	{{0x33,0x28},{0x0A}},
	{{0x33,0x29},{0x06}},
	{{0x33,0x2A},{0x06}},
	{{0x33,0x2B},{0x1C}},
	{{0x33,0x2C},{0x1C}},
	{{0x33,0x2D},{0x00}},
	{{0x33,0x2E},{0x1D}},
	{{0x33,0x2F},{0x1F}},
	{{0x32,0xF6},{0x0F}},
	{{0x32,0xF9},{0x63}},
	{{0x32,0xFA},{0x36}},
	{{0x33,0x25},{0x4A}},
	{{0x33,0x30},{0x00}},
	{{0x33,0x31},{0x0A}},
	{{0x33,0x32},{0xFF}},
	{{0x33,0x38},{0x30}},
	{{0x33,0x39},{0x84}},
	{{0x33,0x3A},{0x48}},
	{{0x33,0x3F},{0x07}},
	{{0x33,0x60},{0x10}},
	{{0x33,0x61},{0x18}},
	{{0x33,0x62},{0x1F}},
	{{0x33,0x63},{0x37}},
	{{0x33,0x64},{0x84}},
	{{0x33,0x65},{0x80}},
	{{0x33,0x66},{0x70}},
	{{0x33,0x67},{0x64}},
	{{0x33,0x68},{0x30}},
	{{0x33,0x69},{0x28}},
	{{0x33,0x6A},{0x20}},
	{{0x33,0x6B},{0x10}},
	{{0x33,0x6C},{0x00}},
	{{0x33,0x6D},{0x20}},
	{{0x33,0x6E},{0x1C}},
	{{0x33,0x6F},{0x18}},
	{{0x33,0x70},{0x10}},
	{{0x33,0x71},{0x38}},
	{{0x33,0x72},{0x3C}},
	{{0x33,0x73},{0x3F}},
	{{0x33,0x74},{0x3F}},
	{{0x33,0x8A},{0x34}},
	{{0x33,0x8B},{0x7F}},
	{{0x33,0x8C},{0x10}},
	{{0x33,0x8D},{0x23}},
	{{0x33,0x8E},{0x7F}},
	{{0x33,0x8F},{0x14}},
	{{0x33,0x75},{0x0A}},
	{{0x33,0x76},{0x0C}},
	{{0x33,0x77},{0x10}},
	{{0x33,0x78},{0x14}},

#if 1 // for 3208B
	{{0x32,0x70},{0x08}},//YW_TUNE
	{{0x32,0x71},{0x16}},
	{{0x32,0x72},{0x22}},
	{{0x32,0x73},{0x37}},
	{{0x32,0x74},{0x49}},
	{{0x32,0x75},{0x5A}},
	{{0x32,0x76},{0x76}},
	{{0x32,0x77},{0x8B}},
	{{0x32,0x78},{0x9F}},
	{{0x32,0x79},{0xB0}},
	{{0x32,0x7A},{0xCD}},
	{{0x32,0x7B},{0xE2}},
	{{0x32,0x7C},{0xEF}},
	{{0x32,0x7D},{0xF9}},
	{{0x32,0x7E},{0xFF}},

  {{0x33,0x02},{0x00}}, //CC CAL 160Tool
  {{0x33,0x03},{0x4C}},
  {{0x33,0x04},{0x00}},
  {{0x33,0x05},{0x96}},
  {{0x33,0x06},{0x00}},
  {{0x33,0x07},{0x1D}},
  {{0x33,0x08},{0x07}},
  {{0x33,0x09},{0xBB}},
  {{0x33,0x0A},{0x06}},
  {{0x33,0x0B},{0xDC}},
  {{0x33,0x0C},{0x01}},
  {{0x33,0x0D},{0x6A}},
  {{0x33,0x0E},{0x01}},
  {{0x33,0x0F},{0x10}},
  {{0x33,0x10},{0x07}},
  {{0x33,0x11},{0x20}},
  {{0x33,0x12},{0x07}},
  {{0x33,0x13},{0xD0}},
#else	// for 3309A
	{{0x32,0x70},{0x10}},
	{{0x32,0x71},{0x1A}},
	{{0x32,0x72},{0x25}},
	{{0x32,0x73},{0x38}},
	{{0x32,0x74},{0x49}},
	{{0x32,0x75},{0x5A}},
	{{0x32,0x76},{0x71}},
	{{0x32,0x77},{0x87}},
	{{0x32,0x78},{0x99}},
	{{0x32,0x79},{0xA6}},
	{{0x32,0x7A},{0xBC}},
	{{0x32,0x7B},{0xCC}},
	{{0x32,0x7C},{0xDD}},
	{{0x32,0x7D},{0xEE}},
	{{0x32,0x7E},{0xFF}},

	{{0x33,0x02},{0x00}},
	{{0x33,0x03},{0x40}},
	{{0x33,0x04},{0x00}},
	{{0x33,0x05},{0x96}},
	{{0x33,0x06},{0x00}},
	{{0x33,0x07},{0x29}},
	{{0x33,0x08},{0x07}},
	{{0x33,0x09},{0xBA}},
	{{0x33,0x0A},{0x06}},
	{{0x33,0x0B},{0xF5}},
	{{0x33,0x0C},{0x01}},
	{{0x33,0x0D},{0x51}},
	{{0x33,0x0E},{0x01}},
	{{0x33,0x0F},{0x30}},
	{{0x33,0x10},{0x07}},
	{{0x33,0x11},{0x16}},
	{{0x33,0x12},{0x07}},
	{{0x33,0x13},{0xBA}},
#endif
	//CY Hsu 20131014 Start
	{{0x32,0x01}, {0x7F}},
	{{0x32,0xB1}, {0x7D}},
	{{0x32,0xB2}, {0x7D}},
	{{0x32,0x57}, {0x60}},
	//CY Hsu 20131014 End
};


//1280x720, 30fps,pclk=80Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_hd720_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5B}}, 
{{0x32,0xC1}, {0x5B}}, 
{{0x32,0xC2}, {0x5B}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x5B}}, 
{{0x32,0xCA}, {0x83}}, 
{{0x32,0xCB}, {0x83}}, 
{{0x32,0xCC}, {0x83}}, 
{{0x32,0xCD}, {0x83}}, 
{{0x32,0xDB}, {0x7B}}, 
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3E}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x0B}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0x04}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x05}}, 
{{0x30,0x07}, {0x03}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x06}}, 
{{0x30,0x0B}, {0x7F}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x05}}, 
{{0x30,0x0F}, {0x00}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x87}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}}, 
};

//1280x720, 25fps,pclk=80Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_hd720_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x60}}, 
{{0x32,0xC2}, {0x60}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x60}}, 
{{0x32,0xCA}, {0x88}}, 
{{0x32,0xCB}, {0x88}}, 
{{0x32,0xCC}, {0x88}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x7B}}, 
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3E}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x0B}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0x04}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x05}}, 
{{0x30,0x07}, {0x03}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x06}}, 
{{0x30,0x0B}, {0x82}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x60}}, 
{{0x30,0x0E}, {0x05}}, 
{{0x30,0x0F}, {0x00}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x87}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}}, 
};

//1280x720, 15fps,pclk=40Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_hd720_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x7B}}, 
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3E}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x0B}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0x04}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x05}}, 
{{0x30,0x07}, {0x03}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x06}}, 
{{0x30,0x0B}, {0x82}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xA0}}, 
{{0x30,0x0E}, {0x05}}, 
{{0x30,0x0F}, {0x00}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x87}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}}, 
};

//1280x720, 5fps,pclk=16Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_hd720_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x83}}, 
{{0x32,0xC1}, {0x84}}, 
{{0x32,0xC2}, {0x84}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x84}}, 
{{0x32,0xCA}, {0xAC}}, 
{{0x32,0xCB}, {0xAC}}, 
{{0x32,0xCC}, {0xAC}}, 
{{0x32,0xCD}, {0xAB}}, 
{{0x32,0xDB}, {0x7B}}, 
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3E}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x0B}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0x04}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x05}}, 
{{0x30,0x07}, {0x03}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x06}}, 
{{0x30,0x0B}, {0x82}}, 
{{0x30,0x0C}, {0x10}}, 
{{0x30,0x0D}, {0xE1}}, 
{{0x30,0x0E}, {0x05}}, 
{{0x30,0x0F}, {0x00}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x87}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}}, 
};


//[YUYV_1280x720_15_25Fps]
static struct regval_list sensor_ae_target0_hd720_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x7B}}, 
{{0x32,0xE0}, {0x05}}, 
{{0x32,0xE1}, {0x00}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x3E}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x0B}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0x04}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x05}}, 
{{0x30,0x07}, {0x03}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x06}}, 
{{0x30,0x0B}, {0x82}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x60}}, 
{{0x30,0x0E}, {0x05}}, 
{{0x30,0x0F}, {0x00}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x87}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}},
};

//[1280x720_30fps_50Hz]
static struct regval_list sensor_ae_target0_hd720_30fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5B}}, 
{{0x32,0xC1}, {0x5B}}, 
{{0x32,0xC2}, {0x5B}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x5B}}, 
{{0x32,0xCA}, {0x83}}, 
{{0x32,0xCB}, {0x83}}, 
{{0x32,0xCC}, {0x83}}, 
{{0x32,0xCD}, {0x83}}, 
{{0x32,0xDB}, {0x7B}}, 
};

//[1280x720_30fps_60Hz]
static struct regval_list sensor_ae_target0_hd720_30fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x60}}, 
{{0x32,0xC2}, {0x60}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB4}}, 
{{0x32,0xC9}, {0x60}}, 
{{0x32,0xCA}, {0x88}}, 
{{0x32,0xCB}, {0x88}}, 
{{0x32,0xCC}, {0x88}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x76}},
};

//1280,720 25FPS [50Hz]
static struct regval_list sensor_ae_target0_hd720_25fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x60}}, 
{{0x32,0xC2}, {0x60}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x60}}, 
{{0x32,0xCA}, {0x88}}, 
{{0x32,0xCB}, {0x88}}, 
{{0x32,0xCC}, {0x88}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x7B}}, 
};

static struct regval_list sensor_ae_target0_hd720_25fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x63}}, 
{{0x32,0xC1}, {0x63}}, 
{{0x32,0xC2}, {0x63}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB4}}, 
{{0x32,0xC9}, {0x63}}, 
{{0x32,0xCA}, {0x8B}}, 
{{0x32,0xCB}, {0x8B}}, 
{{0x32,0xCC}, {0x8B}}, 
{{0x32,0xCD}, {0x8B}}, 
{{0x32,0xDB}, {0x76}}, 
};

//1280,720 15FPS [50Hz]
static struct regval_list sensor_ae_target0_hd720_15fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x7B}}, 
};

static struct regval_list sensor_ae_target0_hd720_15fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB4}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x76}},  
};

//1280,720 5FPS [50Hz]
static struct regval_list sensor_ae_target0_hd720_5fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x83}}, 
{{0x32,0xC1}, {0x84}}, 
{{0x32,0xC2}, {0x84}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x84}}, 
{{0x32,0xCA}, {0xAC}}, 
{{0x32,0xCB}, {0xAC}}, 
{{0x32,0xCC}, {0xAC}}, 
{{0x32,0xCD}, {0xAB}}, 
{{0x32,0xDB}, {0x7B}}, 
};

static struct regval_list sensor_ae_target0_hd720_5fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x88}}, 
{{0x32,0xC1}, {0x88}}, 
{{0x32,0xC2}, {0x88}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB4}}, 
{{0x32,0xC9}, {0x88}}, 
{{0x32,0xCA}, {0xB0}}, 
{{0x32,0xCB}, {0xB0}}, 
{{0x32,0xCC}, {0xB0}}, 
{{0x32,0xCD}, {0xB0}}, 
{{0x32,0xDB}, {0x76}},  
};

//[1280x720_50Hz]
static struct regval_list sensor_ae_target0_hd720_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xD8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x7B}}, 
};

//[1280x720_60Hz]
static struct regval_list sensor_ae_target0_hd720_29fps_60hz_regs[] =  
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB4}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x76}}, 
};

//YUYV_960x720_30.00Fps
static struct regval_list sensor_ae_target0_960hd720_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_960x720_25.00Fps]
static struct regval_list sensor_ae_target0_960hd720_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}}, 
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_960x720_15.00Fps]
static struct regval_list sensor_ae_target0_960hd720_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_960x720_5.00Fps]
static struct regval_list sensor_ae_target0_960hd720_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}},  
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};


//[YUYV_960x720_15_25Fps]
static struct regval_list sensor_ae_target0_960hd720_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}},
};

//[960x720_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_960hd720_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}},
};

//[960x720_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_960hd720_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//[YUYV_880x720_30.00Fps]
static struct regval_list sensor_ae_target0_880hd720_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_880x720_25.00Fps
static struct regval_list sensor_ae_target0_880hd720_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_880x720_15.00Fps
static struct regval_list sensor_ae_target0_880hd720_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}}, 
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_880x720_5.00Fps
static struct regval_list sensor_ae_target0_880hd720_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
//CY Hsu 20131014 Start
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}},  
//CY Hsu 20131014 End
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x3F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_880x720_15_25_Fps]
static struct regval_list sensor_ae_target0_880hd720_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x70}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0xD0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x00}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x00}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x07}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x18}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x00}},
{{0x30,0x60}, {0x01}},
};

//[880x720_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_880hd720_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

static struct regval_list sensor_ae_target0_880hd720_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};


//[YUYV_800x600_15_25_Fps]
static struct regval_list sensor_ae_target0_wvga_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x03}}, 
{{0x32,0xE1}, {0x20}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x58}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x33}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x33}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x48}},
{{0x30,0x60}, {0x01}}, 
};

//[800x600_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_wvga_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[800x600_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_wvga_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//[YUYV_704x576_30.00Fps]
static struct regval_list sensor_ae_target0_4cif_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV, 704x576_25.00Fps
static struct regval_list sensor_ae_target0_4cif_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV, 704x576_15.00Fps
static struct regval_list sensor_ae_target0_4cif_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV, 704x576_5.00Fps
static struct regval_list sensor_ae_target0_4cif_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};
      
    
//[YUYV_704x576_25_30_Fps]
static struct regval_list sensor_ae_target0_4cif_29fps_regs[] =
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0xC0}}, 
{{0x32,0xE2}, {0x02}}, 
{{0x32,0xE3}, {0x40}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x40}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x40}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x07}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x18}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x58}},
{{0x30,0x60}, {0x01}}, 
};

//[704x576_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_4cif_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[704x576_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_4cif_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//640x480, 30fps,pclk=80Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_vga_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//640x480, 25fps,pclk=80Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_vga_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//640x480, 15fps,pclk=40Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_vga_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//640x480, 5fps,pclk=16Mhz(MCLK=24M)  YUV
static struct regval_list sensor_ae_target0_vga_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_640x480_25_30_Fps]
static struct regval_list sensor_ae_target0_vga_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x02}}, 
{{0x32,0xE1}, {0x80}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0xE0}}, 
{{0x32,0xE4}, {0x00}}, 
{{0x32,0xE5}, {0x80}}, 
{{0x32,0xE6}, {0x00}}, 
{{0x32,0xE7}, {0x80}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}},
};

//[640x480_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_vga_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[640x480_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_vga_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}},
};

//[YUYV_352x288_30.00Fps]
static struct regval_list sensor_ae_target0_cif_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_352x288_25.00Fps]
static struct regval_list sensor_ae_target0_cif_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_352x288_15.00Fps]
static struct regval_list sensor_ae_target0_cif_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//YUYV_352x288_5.00Fps
static struct regval_list sensor_ae_target0_cif_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};
      
//[YUYV_352x288_15_25_Fps]
static struct regval_list sensor_ae_target0_cif_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x60}}, 
{{0x32,0xE2}, {0x01}}, 
{{0x32,0xE3}, {0x20}}, 
{{0x32,0xE4}, {0x01}}, 
{{0x32,0xE5}, {0xBC}}, 
{{0x32,0xE6}, {0x01}}, 
{{0x32,0xE7}, {0x82}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}}, 
};

//[352x288_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_cif_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[352x288_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_cif_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};

//320x240
static struct regval_list sensor_ae_target0_qvga_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0},{0x01}},
{{0x32,0xE1},{0x40}},
{{0x32,0xE2},{0x00}},
{{0x32,0xE3},{0xF0}},
{{0x32,0xE4},{0x02}},
{{0x32,0xE5},{0x02}},
{{0x32,0xE6},{0x02}},
{{0x32,0xE7},{0x03}},
{{0x32,0x00},{0x1F}},
//{{0x32,0x01},{0x0F}},
{{0x30,0x28},{0x09}},
{{0x30,0x29},{0x00}},
{{0x30,0x2A},{0x04}},
{{0x30,0x22},{0x24}},
{{0x30,0x23},{0x24}},
{{0x30,0x02},{0x00}},
{{0x30,0x03},{0xA4}},
{{0x30,0x04},{0x00}},
{{0x30,0x05},{0x04}},
{{0x30,0x06},{0x04}},
{{0x30,0x07},{0x63}},
{{0x30,0x08},{0x02}},
{{0x30,0x09},{0xD3}},
{{0x30,0x0A},{0x05}},
{{0x30,0x0B},{0x47}},
{{0x30,0x0C},{0x02}},
{{0x30,0x0D},{0xE4}},
{{0x30,0x0E},{0x03}},
{{0x30,0x0F},{0xC0}},
{{0x30,0x10},{0x02}},
{{0x30,0x11},{0xD0}},
{{0x32,0xB8},{0x3F}},
{{0x32,0xB9},{0x31}},
{{0x32,0xBB},{0x17}},
{{0x32,0xBC},{0x38}},
{{0x32,0xBD},{0x3C}},
{{0x32,0xBE},{0x34}},
//{{0x32,0x01},{0x7F}},
{{0x30,0x21},{0x06}},
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60},{0x01}},
};

static struct regval_list sensor_ae_target0_qvga_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

static struct regval_list sensor_ae_target0_qvga_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//320x240
static struct regval_list sensor_ae_target0_qvga_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}},
};

//[YUYV_320x240_15_25_Fps]
static struct regval_list sensor_ae_target0_qvga_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x01}}, 
{{0x32,0xE1}, {0x40}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0xF0}}, 
{{0x32,0xE4}, {0x02}}, 
{{0x32,0xE5}, {0x02}}, 
{{0x32,0xE6}, {0x02}}, 
{{0x32,0xE7}, {0x03}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}}, 
};

//[320x240_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_qvga_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[320x240_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_qvga_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x20}}, 
{{0x32,0xC5}, {0x20}}, 
{{0x32,0xC6}, {0x20}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x90}}, 
{{0x32,0xCB}, {0x90}}, 
{{0x32,0xCC}, {0x90}}, 
{{0x32,0xCD}, {0x90}}, 
{{0x32,0xDB}, {0x73}}, 
};

//[YUYV_176x144_30Fps]
static struct regval_list sensor_ae_target0_qcif_30fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE4}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};


//[YUYV_176x144_25Fps]
static struct regval_list sensor_ae_target0_qcif_25fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x03}}, 
{{0x30,0x0D}, {0x78}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_176x144_15.00]
static struct regval_list sensor_ae_target0_qcif_15fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x05}}, 
{{0x30,0x0D}, {0xC8}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_176x144_5.00]
static struct regval_list sensor_ae_target0_qcif_5fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},

{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x07}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x1F}}, 
//{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x09}}, 
{{0x30,0x29}, {0x00}}, 
{{0x30,0x2A}, {0x04}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xCC}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x3B}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x47}}, 
{{0x30,0x0C}, {0x11}}, 
{{0x30,0x0D}, {0x59}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0x70}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
//{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0xB2}},
{{0x30,0x60}, {0x01}}, 
};

//[YUYV_176x144_25_30_Fps]
static struct regval_list sensor_ae_target0_qcif_29fps_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
{{0x32,0xE0}, {0x00}}, 
{{0x32,0xE1}, {0xB0}}, 
{{0x32,0xE2}, {0x00}}, 
{{0x32,0xE3}, {0x90}}, 
{{0x32,0xE4}, {0x04}}, 
{{0x32,0xE5}, {0x7C}}, 
{{0x32,0xE6}, {0x04}}, 
{{0x32,0xE7}, {0x08}}, 
{{0x32,0x00}, {0x1F}}, 
{{0x32,0x01}, {0x0F}}, 
{{0x30,0x28}, {0x18}}, 
{{0x30,0x29}, {0x12}}, 
{{0x30,0x2A}, {0x00}}, 
{{0x30,0x22}, {0x24}}, 
{{0x30,0x23}, {0x24}}, 
{{0x30,0x02}, {0x00}}, 
{{0x30,0x03}, {0xA4}}, 
{{0x30,0x04}, {0x00}}, 
{{0x30,0x05}, {0x04}}, 
{{0x30,0x06}, {0x04}}, 
{{0x30,0x07}, {0x63}}, 
{{0x30,0x08}, {0x02}}, 
{{0x30,0x09}, {0xD3}}, 
{{0x30,0x0A}, {0x05}}, 
{{0x30,0x0B}, {0x4E}}, 
{{0x30,0x0C}, {0x02}}, 
{{0x30,0x0D}, {0xE0}}, 
{{0x30,0x0E}, {0x03}}, 
{{0x30,0x0F}, {0xC0}}, 
{{0x30,0x10}, {0x02}}, 
{{0x30,0x11}, {0xD0}}, 
{{0x32,0xB8}, {0x3F}}, 
{{0x32,0xB9}, {0x31}}, 
{{0x32,0xBB}, {0x17}}, 
{{0x32,0xBC}, {0x38}}, 
{{0x32,0xBD}, {0x3C}}, 
{{0x32,0xBE}, {0x34}}, 
{{0x32,0x01}, {0x7F}}, 
{{0x30,0x21}, {0x06}}, 
{{0x32,0x0A}, {0x72}},
{{0x30,0x60}, {0x01}}, 
};

//[176x144_15_25_Fps_50Hz]
static struct regval_list sensor_ae_target0_qcif_29fps_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x6A}}, 
{{0x32,0xC1}, {0x6A}}, 
{{0x32,0xC2}, {0x6A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB8}}, 
{{0x32,0xC9}, {0x6A}}, 
{{0x32,0xCA}, {0x92}}, 
{{0x32,0xCB}, {0x92}}, 
{{0x32,0xCC}, {0x92}}, 
{{0x32,0xCD}, {0x92}}, 
{{0x32,0xDB}, {0x77}}, 
};

//[176x144_15_25_Fps_60Hz]
static struct regval_list sensor_ae_target0_qcif_29fps_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x70}}, 
{{0x32,0xC1}, {0x70}}, 
{{0x32,0xC2}, {0x70}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0x99}}, 
{{0x32,0xC9}, {0x70}}, 
{{0x32,0xCA}, {0x98}}, 
{{0x32,0xCB}, {0x98}}, 
{{0x32,0xCC}, {0x98}}, 
{{0x32,0xCD}, {0x98}}, 
{{0x32,0xDB}, {0x73}}, 
};

//960x720,880x720,800x600,704x576,640x480,352x288,320x240,176x144 Anti Flicker]
//[50Hz]
static struct regval_list sensor_ae_target0_default_50hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x5A}}, 
{{0x32,0xC1}, {0x5A}}, 
{{0x32,0xC2}, {0x5A}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xDE}}, 
{{0x32,0xC9}, {0x5A}}, 
{{0x32,0xCA}, {0x82}}, 
{{0x32,0xCB}, {0x82}}, 
{{0x32,0xCC}, {0x82}}, 
{{0x32,0xCD}, {0x82}}, 
{{0x32,0xDB}, {0x7B}},
};

//[60Hz]
static struct regval_list sensor_ae_target0_default_60hz_regs[] = 
{
{{0x32,0xBF}, {0x60}}, 
{{0x32,0xC0}, {0x60}}, 
{{0x32,0xC1}, {0x5F}}, 
{{0x32,0xC2}, {0x5F}}, 
{{0x32,0xC3}, {0x00}}, 
{{0x32,0xC4}, {0x27}}, 
{{0x32,0xC5}, {0x28}}, 
{{0x32,0xC6}, {0x28}}, 
{{0x32,0xC7}, {0x00}}, 
{{0x32,0xC8}, {0xB9}}, 
{{0x32,0xC9}, {0x5F}}, 
{{0x32,0xCA}, {0x87}}, 
{{0x32,0xCB}, {0x87}}, 
{{0x32,0xCC}, {0x87}}, 
{{0x32,0xCD}, {0x88}}, 
{{0x32,0xDB}, {0x77}}, 
};


/*
 * Here we'll try to encapsulate the changes for just the output
 * video format.
 * 
 */


static struct regval_list sensor_fmt_yuv422_yuyv[] = {
	//YCbYCr
	{{0x32,0xF0}, {0x01}},
};


static struct regval_list sensor_fmt_yuv422_yvyu[] = {
	//YCrYCb
	{{0x32,0xF0}, {0x03}},
};

static struct regval_list sensor_fmt_yuv422_vyuy[] = {
	//CrYCbY
	{{0x32,0xF0}, {0x02}},
};

static struct regval_list sensor_fmt_yuv422_uyvy[] = {
	//CbYCrY
	{{0x32,0xF0}, {0x00}},
};

/*
 * Low-level register I/O.
 *
 */


/*
 * On most platforms, we'd rather do straight i2c I/O.
 */
static int sensor_read(struct v4l2_subdev *sd, unsigned char *reg,
		unsigned char *value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	u8 data[REG_STEP];
	struct i2c_msg msg;
	int ret,i;
	
	for(i = 0; i < REG_ADDR_STEP; i++)
		data[i] = reg[i];
	
	data[REG_ADDR_STEP] = 0xff;
	/*
	 * Send out the register address...
	 */
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_ADDR_STEP;
	msg.buf = data;
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret < 0) {
		csi_dev_err("Error %d on %d register write\n", ret, msg.addr);
		return ret;
	}
	/*
	 * ...then read back the result.
	 */
	
	msg.flags = I2C_M_RD;
	msg.len = REG_DATA_STEP;
	msg.buf = &data[REG_ADDR_STEP];
	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret >= 0) {
		for(i = 0; i < REG_DATA_STEP; i++)
			value[i] = data[i+REG_ADDR_STEP];
		ret = 0;
	}
	else {
		csi_dev_err("Error %d on register read\n", ret);
	}
	return ret;
}


static int sensor_write(struct v4l2_subdev *sd, unsigned char *reg,
		unsigned char *value)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct i2c_msg msg;
	unsigned char data[REG_STEP];
	int ret,i;
	
	for(i = 0; i < REG_ADDR_STEP; i++)
			data[i] = reg[i];
	for(i = REG_ADDR_STEP; i < REG_STEP; i++)
			data[i] = value[i-REG_ADDR_STEP];
	
//	for(i = 0; i < REG_STEP; i++)
//		printk("data[%x]=%x\n",i,data[i]);
			
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_STEP;
	msg.buf = data;

	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret > 0) {
		ret = 0;
	}
	else if (ret < 0) {
		csi_dev_err("sensor_write error!\n");
	}
	return ret;
}

static int sensor_read_client(struct i2c_client *client, unsigned char *reg,
		unsigned char *value)
{
	u8 data[REG_STEP];
	struct i2c_msg msg;
	int ret,i;
	
	for(i = 0; i < REG_ADDR_STEP; i++)
		data[i] = reg[i];
	
	data[REG_ADDR_STEP] = 0xff;
	/*
	 * Send out the register address...
	 */
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_ADDR_STEP;
	msg.buf = data;
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret < 0) {
		csi_dev_err("Error %d on %d register write\n", ret, msg.addr);
		return ret;
	}
	/*
	 * ...then read back the result.
	 */
	
	msg.flags = I2C_M_RD;
	msg.len = REG_DATA_STEP;
	msg.buf = &data[REG_ADDR_STEP];
	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret >= 0) {
		for(i = 0; i < REG_DATA_STEP; i++)
			value[i] = data[i+REG_ADDR_STEP];
		ret = 0;
	}
	else {
		csi_dev_err("Error %d on register read\n", ret);
	}
	return ret;
}


static int sensor_write_client(struct i2c_client *client, unsigned char *reg,
		unsigned char *value)
{
	struct i2c_msg msg;
	unsigned char data[REG_STEP];
	int ret,i;
	
	for(i = 0; i < REG_ADDR_STEP; i++)
			data[i] = reg[i];
	for(i = REG_ADDR_STEP; i < REG_STEP; i++)
			data[i] = value[i-REG_ADDR_STEP];
	
//	for(i = 0; i < REG_STEP; i++)
//		printk("data[%x]=%x\n",i,data[i]);
			
	msg.addr = client->addr;
	msg.flags = 0;
	msg.len = REG_STEP;
	msg.buf = data;

	
	ret = i2c_transfer(client->adapter, &msg, 1);
	if (ret > 0) {
		ret = 0;
	}
	else if (ret < 0) {
		csi_dev_err("sensor_write error!\n");
	}
	return ret;
}

static int sensor_write_array_client(struct i2c_client *client, struct regval_list *vals , uint size)
{
	int i,ret;
	
	if (size == 0)
		return -EINVAL;
	
	for(i = 0; i < size ; i++)
	{
		if(vals->reg_num[0] == 0xff) {
			mdelay(vals->value[0]);
		} else {
			ret = sensor_write_client(client, vals->reg_num, vals->value);
			if (ret < 0) {
				csi_dev_err("sensor_write_err!\n");
				return ret;
			}
		}
		
		vals++;
	}

	return 0;
}

/*
 * Write a list of register settings;
 */
static int sensor_write_array(struct v4l2_subdev *sd, struct regval_list *vals , uint size)
{
	int i,ret;
	
	if (size == 0)
		return -EINVAL;
	
	for(i = 0; i < size ; i++)
	{
		if(vals->reg_num[0] == 0xff) {
			mdelay(vals->value[0]);
		} else {
			ret = sensor_write(sd, vals->reg_num, vals->value);
			if (ret < 0) {
				csi_dev_err("sensor_write_err!\n");
				return ret;
			}
		}
		
		vals++;
	}

	return 0;
}

/*
 * CSI GPIO control
 */
static void csi_gpio_write(struct v4l2_subdev *sd, struct gpio_config *gpio, int level)
{
//	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
  if(gpio->gpio==GPIO_INDEX_INVALID)
  {
    return;
  }
  
	if(gpio->mul_sel==1)
	{
	  gpio_direction_output(gpio->gpio, level);
	  gpio->data=level;
	} else {
	  csi_dev_dbg("gpio is not in output function\n");
	}
}

static void csi_gpio_set_status(struct v4l2_subdev *sd, struct gpio_config *gpio, int status)
{
//	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
	if(1 == status && gpio->gpio != GPIO_INDEX_INVALID) {  /* output */
		if(0 != gpio_direction_output(gpio->gpio, gpio->data))
			csi_dev_dbg("gpio_direction_output failed\n");
		else {
		  csi_dev_dbg("gpio_direction_output gpio[%d]=%d\n",gpio->gpio, gpio->data);
			gpio->mul_sel=status;
		}
	} else if(0 == status && gpio->gpio != GPIO_INDEX_INVALID) {  /* input */
	  if(0 != gpio_direction_input(gpio->gpio) )
	    csi_dev_dbg("gpio_direction_input failed\n");
	  else
	    gpio->mul_sel=status;
	}
}


/*
 * Stuff that knows about the sensor.
 */
 
static int sensor_power(struct v4l2_subdev *sd, int on)
{
	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	int ret;
	
  //make sure that no device can access i2c bus during sensor initial or power down
  //when using i2c_lock_adpater function, the following codes must not access i2c bus before calling i2c_unlock_adapter
  i2c_lock_adapter(client->adapter);
  //insure that clk_disable() and clk_enable() are called in pair 
  //when calling CSI_SUBDEV_STBY_ON/OFF and CSI_SUBDEV_PWR_ON/OFF
  switch(on)
	{
		case CSI_SUBDEV_STBY_ON:
			csi_dev_dbg("CSI_SUBDEV_STBY_ON\n");
			//reset off io
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			//disable oe
			csi_dev_print("disalbe oe!\n");
			ret = sensor_write_array(sd,sensor_oe_disable,ARRAY_SIZE(sensor_oe_disable));
			if(ret < 0)
				csi_dev_err("sensor_oe_disable error\n");
			
			//standby on io
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);
			mdelay(1);
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_OFF);
			mdelay(1);
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);
			mdelay(1);
			//inactive mclk after stadby in
			clk_disable(dev->csi_module_clk);
			//reset on io
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			break;
		case CSI_SUBDEV_STBY_OFF:
			csi_dev_dbg("CSI_SUBDEV_STBY_OFF\n");
			//active mclk before stadby out
			clk_enable(dev->csi_module_clk);
			mdelay(1);
			//standby off io
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_OFF);
			mdelay(1);
			//reset off io
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			break;
		case CSI_SUBDEV_PWR_ON:
			csi_dev_dbg("CSI_SUBDEV_PWR_ON\n");
			//power on reset
			csi_gpio_set_status(sd,&dev->standby_io,1);//set the gpio to output
			csi_gpio_set_status(sd,&dev->reset_io,1);//set the gpio to output
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);
			//reset on io
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			//active mclk before power on
			clk_enable(dev->csi_module_clk);
			mdelay(1);
			//power supply
			csi_gpio_write(sd,&dev->power_io,CSI_PWR_ON);
			mdelay(1);
			if(dev->dvdd) {
				regulator_enable(dev->dvdd);
				mdelay(1);
			}
			if(dev->avdd) {
				regulator_enable(dev->avdd);
				mdelay(1);
			}
			if(dev->iovdd) {
				regulator_enable(dev->iovdd);
				mdelay(1);
			}
			//standby off io
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_OFF);
			mdelay(1);
			//reset after power on
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			break;
		case CSI_SUBDEV_PWR_OFF:
			csi_dev_dbg("CSI_SUBDEV_PWR_OFF\n");
			//standby and reset io
			csi_gpio_write(sd,&dev->standby_io,CSI_STBY_ON);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			//power supply off
			if(dev->iovdd) {
				regulator_disable(dev->iovdd);
				mdelay(1);
			}
			if(dev->avdd) {
				regulator_disable(dev->avdd);
				mdelay(1);
			}
			if(dev->dvdd) {
				regulator_disable(dev->dvdd);
				mdelay(1);	
			}
			csi_gpio_write(sd,&dev->power_io,CSI_PWR_OFF);
			mdelay(1);
			//inactive mclk after power off
			clk_disable(dev->csi_module_clk);
			//set the io to hi-z
			csi_gpio_set_status(sd,&dev->reset_io,0);//set the gpio to input
			csi_gpio_set_status(sd,&dev->standby_io,0);//set the gpio to input
			break;
		default:
			return -EINVAL;
	}		

	//remember to unlock i2c adapter, so the device can access the i2c bus again
	i2c_unlock_adapter(client->adapter);	
	return 0;
}
 
static int sensor_reset(struct v4l2_subdev *sd, u32 val)
{
	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);

	switch(val)
	{
		case CSI_SUBDEV_RST_OFF:
			csi_dev_dbg("CSI_SUBDEV_RST_OFF\n");
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			break;
		case CSI_SUBDEV_RST_ON:
			csi_dev_dbg("CSI_SUBDEV_RST_ON\n");
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			break;
		case CSI_SUBDEV_RST_PUL:
			csi_dev_dbg("CSI_SUBDEV_RST_PUL\n");
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_ON);
			mdelay(1);
			csi_gpio_write(sd,&dev->reset_io,CSI_RST_OFF);
			mdelay(1);
			break;
		default:
			return -EINVAL;
	}
		
	return 0;
}

static int sensor_test(struct v4l2_subdev *sd)
{
    struct i2c_client *client = v4l2_get_subdevdata(sd);
    unsigned char reg[REG_ADDR_STEP], value[REG_DATA_STEP];
    int ret;
	
    if(NULL == client)
        printk("sensor_test failed to get i2c_client\n");
    
    reg[0] = 0x30;
    reg[1] = 0x69;
    ret = sensor_read_client(client, reg, value);
	if (ret < 0)
	{
		csi_dev_err("sensor_test sensor_read err!\n");		
	}
	else
	{
		csi_dev_err("sensor_test sensor read value is %d\n", value[0]);
	}
	
}

static int sensor_detect(struct v4l2_subdev *sd)
{
	int ret;
	struct regval_list regs;
	
	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x00;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_detect!\n");
		return ret;
	}
	
	if(regs.value[0] != 0x14)
		return -ENODEV;
	
	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x01;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_detect!\n");
		return ret;
	}
	
	if(regs.value[0] != 0x10)
		return -ENODEV;
	
	return 0;
}

static int sensor_init(struct v4l2_subdev *sd, u32 val)
{
	int ret;
	csi_dev_dbg("sensor_init\n");
	//dump_stack();

	/*Make sure it is a target sensor*/
	ret = sensor_detect(sd);
	if (ret) {
		csi_dev_err("chip found is not an target chip.\n");
		return ret;
	}
	return sensor_write_array(sd, sensor_ae_target0_default_regs , ARRAY_SIZE(sensor_ae_target0_default_regs));
}

static long sensor_ioctl(struct v4l2_subdev *sd, unsigned int cmd, void *arg)
{
	int ret=0;

	switch(cmd){
		case CSI_SUBDEV_CMD_GET_INFO: 
		{
			struct sensor_info *info = to_state(sd);
			__csi_subdev_info_t *ccm_info = arg;
			
			csi_dev_dbg("CSI_SUBDEV_CMD_GET_INFO\n");
			
			ccm_info->mclk 	=	info->ccm_info->mclk ;
			ccm_info->vref 	=	info->ccm_info->vref ;
			ccm_info->href 	=	info->ccm_info->href ;
			ccm_info->clock	=	info->ccm_info->clock;
			ccm_info->iocfg	=	info->ccm_info->iocfg;

			csi_dev_dbg("ccm_info.mclk=%x\n ",info->ccm_info->mclk);
			csi_dev_dbg("ccm_info.vref=%x\n ",info->ccm_info->vref);
			csi_dev_dbg("ccm_info.href=%x\n ",info->ccm_info->href);
			csi_dev_dbg("ccm_info.clock=%x\n ",info->ccm_info->clock);
			csi_dev_dbg("ccm_info.iocfg=%x\n ",info->ccm_info->iocfg);
			break;
		}
		case CSI_SUBDEV_CMD_SET_INFO:
		{
			struct sensor_info *info = to_state(sd);
			__csi_subdev_info_t *ccm_info = arg;
			
			csi_dev_dbg("CSI_SUBDEV_CMD_SET_INFO\n");
		
			info->ccm_info->mclk 	=	ccm_info->mclk 	;
			info->ccm_info->vref 	=	ccm_info->vref 	;
			info->ccm_info->href 	=	ccm_info->href 	;
			info->ccm_info->clock	=	ccm_info->clock	;
			info->ccm_info->iocfg	=	ccm_info->iocfg	;
			
			csi_dev_dbg("ccm_info.mclk=%x\n ",info->ccm_info->mclk);
			csi_dev_dbg("ccm_info.vref=%x\n ",info->ccm_info->vref);
			csi_dev_dbg("ccm_info.href=%x\n ",info->ccm_info->href);
			csi_dev_dbg("ccm_info.clock=%x\n ",info->ccm_info->clock);
			csi_dev_dbg("ccm_info.iocfg=%x\n ",info->ccm_info->iocfg);
			break;
		}
		default:
			return -EINVAL;
	}		
	
	return ret;
}

/*
 * Store information about the video data format. 
 */
static struct sensor_format_struct {
	__u8 *desc;
	//__u32 pixelformat;
	enum v4l2_mbus_pixelcode mbus_code;//linux-3.0
	struct regval_list *regs;
	int	regs_size;
	int bpp;   /* Bytes per pixel */
} sensor_formats[] = {
	{
		.desc		= "YUYV 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_YUYV8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_yuyv,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_yuyv),
		.bpp		= 2,
	},
	{
		.desc		= "YVYU 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_YVYU8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_yvyu,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_yvyu),
		.bpp		= 2,
	},
	{
		.desc		= "UYVY 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_UYVY8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_uyvy,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_uyvy),
		.bpp		= 2,
	},
	{
		.desc		= "VYUY 4:2:2",
		.mbus_code	= V4L2_MBUS_FMT_VYUY8_2X8,//linux-3.0
		.regs 		= sensor_fmt_yuv422_vyuy,
		.regs_size = ARRAY_SIZE(sensor_fmt_yuv422_vyuy),
		.bpp		= 2,
	},
//	{
//		.desc		= "Raw RGB Bayer",
//		.mbus_code	= V4L2_MBUS_FMT_SBGGR8_1X8,//linux-3.0
//		.regs 		= sensor_fmt_raw,
//		.regs_size = ARRAY_SIZE(sensor_fmt_raw),
//		.bpp		= 1
//	},
};
#define N_FMTS ARRAY_SIZE(sensor_formats)

/*
 * Then there is the issue of window sizes.  Try to capture the info here.
 * Must arrange the size from high resolution to low
 */
static struct sensor_win_size {
	int	width;
	int	height;
	int	hstart;		/* Start/stop values for the camera.  Note */
	int	hstop;		/* that they do not always make complete */
	int	vstart;		/* sense to humans, but evidently the sensor */
	int	vstop;		/* will do the right thing... */
	struct regval_list *regs; /* Regs to tweak */
	int regs_size;
	int (*set_size) (struct v4l2_subdev *sd);
/* h/vref stuff */
} sensor_win_sizes[] = {

	/**************************AE target mode hangzhou**************************/
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_hd720_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_hd720_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_hd720_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_hd720_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_hd720_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_hd720_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_hd720_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_hd720_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_hd720_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_hd720_5fps_regs),
		.set_size	= NULL,
	},	
	/* 960 720P */
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_960hd720_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_960hd720_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_960hd720_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_960hd720_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_960hd720_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_960hd720_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_960hd720_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_960hd720_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_960hd720_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_960hd720_5fps_regs),
		.set_size	= NULL,
	},
	/* 880 720P */
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_880hd720_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_880hd720_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_880hd720_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_880hd720_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_880hd720_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_880hd720_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_880hd720_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_880hd720_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target0_880hd720_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_880hd720_5fps_regs),
		.set_size	= NULL,
	},
	/* 4CIF */
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target0_4cif_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_4cif_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target0_4cif_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_4cif_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target0_4cif_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_4cif_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target0_4cif_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_4cif_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target0_4cif_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_4cif_5fps_regs),
		.set_size	= NULL,
	},
	/* VGA */
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target0_vga_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_vga_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target0_vga_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_vga_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target0_vga_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_vga_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target0_vga_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_vga_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target0_vga_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_vga_5fps_regs),
		.set_size	= NULL,
	},
	/* CIF */
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target0_cif_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_cif_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target0_cif_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_cif_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target0_cif_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_cif_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target0_cif_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_cif_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target0_cif_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_cif_5fps_regs),
		.set_size	= NULL,
	},
	/* QVGA */
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target0_qvga_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qvga_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target0_qvga_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qvga_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target0_qvga_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qvga_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target0_qvga_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qvga_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target0_qvga_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qvga_5fps_regs),
		.set_size	= NULL,
	},
	/* QCIF */
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target0_qcif_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qcif_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target0_qcif_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qcif_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target0_qcif_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qcif_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target0_qcif_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qcif_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target0_qcif_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target0_qcif_5fps_regs),
		.set_size	= NULL,
	},	

	/*************************shenzheng drakness*********************************/
	/* 720p */
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_hd720_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_hd720_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_hd720_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_hd720_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_hd720_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_hd720_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_hd720_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_hd720_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_hd720_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_hd720_5fps_regs),
		.set_size	= NULL,
	},	
	/* 960 720P */
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_960hd720_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_960hd720_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_960hd720_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_960hd720_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_960hd720_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_960hd720_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_960hd720_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_960hd720_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_960_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_960hd720_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_960hd720_5fps_regs),
		.set_size	= NULL,
	},
	/* 880 720P */
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_880hd720_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_880hd720_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_880hd720_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_880hd720_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_880hd720_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_880hd720_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_880hd720_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_880hd720_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= HD720_880_WIDTH,
		.height		= HD720_HEIGHT,
		.regs 		= sensor_ae_target1_880hd720_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_880hd720_5fps_regs),
		.set_size	= NULL,
	},
	/* 4CIF */
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target1_4cif_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_4cif_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target1_4cif_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_4cif_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target1_4cif_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_4cif_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target1_4cif_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_4cif_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF4_WIDTH,
		.height		= CIF4_HEIGHT,
		.regs 		= sensor_ae_target1_4cif_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_4cif_5fps_regs),
		.set_size	= NULL,
	},
	/* VGA */
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target1_vga_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_vga_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target1_vga_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_vga_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target1_vga_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_vga_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target1_vga_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_vga_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= VGA_WIDTH,
		.height		= VGA_HEIGHT,
		.regs 		= sensor_ae_target1_vga_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_vga_5fps_regs),
		.set_size	= NULL,
	},
	/* CIF */
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target1_cif_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_cif_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target1_cif_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_cif_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target1_cif_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_cif_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target1_cif_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_cif_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= CIF_WIDTH,
		.height		= CIF_HEIGHT,
		.regs 		= sensor_ae_target1_cif_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_cif_5fps_regs),
		.set_size	= NULL,
	},
	/* QVGA */
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target1_qvga_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qvga_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target1_qvga_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qvga_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target1_qvga_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qvga_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target1_qvga_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qvga_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QVGA_WIDTH,
		.height		= QVGA_HEIGHT,
		.regs 		= sensor_ae_target1_qvga_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qvga_5fps_regs),
		.set_size	= NULL,
	},
	/* QCIF */
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target1_qcif_25fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qcif_25fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target1_qcif_30fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qcif_30fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target1_qcif_29fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qcif_29fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target1_qcif_15fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qcif_15fps_regs),
		.set_size	= NULL,
	},
	{
		.width		= QCIF_WIDTH,
		.height		= QCIF_HEIGHT,
		.regs 		= sensor_ae_target1_qcif_5fps_regs,
		.regs_size	= ARRAY_SIZE(sensor_ae_target1_qcif_5fps_regs),
		.set_size	= NULL,
	},
};

#define N_WIN_SIZES ((ARRAY_SIZE(sensor_win_sizes))>>1) //now there are two mode, so right shift 1 means divide two
static struct sensor_win_size* current_winsize = sensor_win_sizes;
  
static int sensor_enum_fmt(struct v4l2_subdev *sd, unsigned index,
                 enum v4l2_mbus_pixelcode *code)//linux-3.0
{
//	struct sensor_format_struct *ofmt;

	if (index >= N_FMTS)//linux-3.0
		return -EINVAL;

	*code = sensor_formats[index].mbus_code;//linux-3.0
//	ofmt = sensor_formats + fmt->index;
//	fmt->flags = 0;
//	strcpy(fmt->description, ofmt->desc);
//	fmt->pixelformat = ofmt->pixelformat;
	return 0;
}


static int sensor_try_fmt_internal(struct v4l2_subdev *sd,
		//struct v4l2_format *fmt,
		struct v4l2_mbus_framefmt *fmt,//linux-3.0
		struct sensor_format_struct **ret_fmt,
		struct sensor_win_size **ret_wsize)
{
	int index;
	struct sensor_win_size *wsize, *srcwsize;
	struct sensor_info *info = to_state(sd);
//	struct v4l2_pix_format *pix = &fmt->fmt.pix;//linux-3.0
	csi_dev_dbg("sensor_try_fmt_internal\n");
	for (index = 0; index < N_FMTS; index++)
		if (sensor_formats[index].mbus_code == fmt->code)//linux-3.0
			break;
	
	if (index >= N_FMTS) {
		/* default to first format */
		index = 0;
		fmt->code = sensor_formats[0].mbus_code;//linux-3.0
	}
	
	if (ret_fmt != NULL)
		*ret_fmt = sensor_formats + index;
		
	/*
	 * Fields: the sensor devices claim to be progressive.
	 */
	fmt->field = V4L2_FIELD_NONE;//linux-3.0
	
	
	/*
	 * Round requested image size down to the nearest
	 * we support, but not below the smallest.
	 */
	srcwsize = sensor_win_sizes+info->aetarget*N_WIN_SIZES;	
	for (wsize = srcwsize; wsize < srcwsize + N_WIN_SIZES; wsize++)
		if (fmt->width >= wsize->width && fmt->height >= wsize->height)//linux-3.0
			break;
	
	if (wsize >= srcwsize + N_WIN_SIZES)
		wsize--;   /* Take the smallest one */
	if (ret_wsize != NULL)
		*ret_wsize = wsize;
	/*
	 * Note the size we'll actually handle.
	 */
	fmt->width = wsize->width;//linux-3.0
	fmt->height = wsize->height;//linux-3.0
	//pix->bytesperline = pix->width*sensor_formats[index].bpp;//linux-3.0
	//pix->sizeimage = pix->height*pix->bytesperline;//linux-3.0
	
	return 0;
}

static int sensor_try_fmt(struct v4l2_subdev *sd, 
             struct v4l2_mbus_framefmt *fmt)//linux-3.0
{
	return sensor_try_fmt_internal(sd, fmt, NULL, NULL);
}

/*
 * Set a format.
 */
static int sensor_s_fmt(struct v4l2_subdev *sd, 
             struct v4l2_mbus_framefmt *fmt)//linux-3.0
{
	int ret;
	struct sensor_format_struct *sensor_fmt;
	struct sensor_win_size *wsize;
	struct sensor_info *info = to_state(sd);
	csi_dev_dbg("sensor_s_fmt\n");
	ret = sensor_try_fmt_internal(sd, fmt, &sensor_fmt, &wsize);
	if (ret)
		return ret;
		
	sensor_write_array(sd, sensor_fmt->regs , sensor_fmt->regs_size);
	
	ret = 0;
	if (wsize->regs)
	{
		ret = sensor_write_array(sd, wsize->regs , wsize->regs_size);
		if (ret < 0)
			return ret;
	}
		
	if (wsize->set_size)
	{
		ret = wsize->set_size(sd);
		if (ret < 0)
			return ret;
	}
	current_winsize = wsize;
	printk(KERN_ERR "sensor_s_fmt width:%d,height:%d\n", wsize->width, wsize->height);
	info->fmt = sensor_fmt;
	info->width = wsize->width;
	info->height = wsize->height;  

#if 0
	wsize += N_FREQUENCY_INDEX;
	if (wsize->regs)
	{
		ret = sensor_write_array(sd, wsize->regs , wsize->regs_size);
		if (ret < 0)
			return ret;
	}
#endif
	return 0;
}

/*
 * Implement G/S_PARM.  There is a "high quality" mode we could try
 * to do someday; for now, we just do the frame rate tweak.
 */
static int sensor_g_parm(struct v4l2_subdev *sd, struct v4l2_streamparm *parms)
{
	struct v4l2_captureparm *cp = &parms->parm.capture;
	//struct sensor_info *info = to_state(sd);

	if (parms->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	memset(cp, 0, sizeof(struct v4l2_captureparm));
	cp->capability = V4L2_CAP_TIMEPERFRAME;
	cp->timeperframe.numerator = 1;
	cp->timeperframe.denominator = SENSOR_FRAME_RATE;
	
	return 0;
}

static int sensor_s_parm(struct v4l2_subdev *sd, struct v4l2_streamparm *parms)
{
	struct v4l2_captureparm *cp = &parms->parm.capture;
	struct v4l2_fract *tpf = &cp->timeperframe;
	struct sensor_win_size *wsize = current_winsize;
	struct sensor_info *info = to_state(sd);
	unsigned short division = 0;

	printk(KERN_ERR "enter sensor_s_parm\n");
	if (parms->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	if (tpf->numerator == 0 || tpf->denominator == 0)
	{
		division = 1;
	}
	else
	{
		division = tpf->numerator*(SENSOR_FRAME_RATE/tpf->denominator);
	}
	
	printk(KERN_ERR "numerator:%d, denominator:%d, div:%d\n", tpf->numerator, tpf->denominator, division);

	if (division == 0)
		division = 1;
	else if (division > 6)
		division = 6;
	info->clkrc = (info->clkrc & 0x80) | division;

	if (tpf->numerator == 0)
		tpf->numerator = 1;
	
	//sensor_write(sd, REG_CLKRC, info->clkrc);
	//tpf->denominator = SENSOR_FRAME_RATE/division;
	
	printk(KERN_ERR "enter sensor_s_parm fps:%d, width:%d, height:%d\n", tpf->denominator, wsize->width, wsize->height);
	printk(KERN_ERR "wsize is %d offset\n", wsize-sensor_win_sizes);
	switch(tpf->denominator)
	{
	case 30:
		printk(KERN_ERR "30fps");
		wsize += 1;
		break;
	case 29:
		printk(KERN_ERR "29fps");
		wsize += 2;
		break;
	case 15:
		printk(KERN_ERR "15fps");
		wsize += 3;
		break;
	case 5:
		printk(KERN_ERR "5fps");
		wsize += 4;
		break;
	default:
		return 0;
	}
	sensor_write_array(sd, wsize->regs, wsize->regs_size); 

	return 0;
}


/* 
 * Code for dealing with controls.
 * fill with different sensor module
 * different sensor module has different settings here
 * if not support the follow function ,retrun -EINVAL
 */

/* *********************************************begin of ******************************************** */
static int sensor_queryctrl(struct v4l2_subdev *sd,
		struct v4l2_queryctrl *qc)
{
	/* Fill in min, max, step and default value for these controls. */
	/* see include/linux/videodev2.h for details */
	/* see sensor_s_parm and sensor_g_parm for the meaning of value */
	
	switch (qc->id) {
	case V4L2_CID_BRIGHTNESS:
		return  v4l2_ctrl_query_fill(qc, -4, 4, 1, 1);
//	case V4L2_CID_CONTRAST:
//		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 1);
	case V4L2_CID_SATURATION:
		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 1);
//	case V4L2_CID_HUE:
//		return v4l2_ctrl_query_fill(qc, -180, 180, 5, 0);
	case V4L2_CID_VFLIP:
	case V4L2_CID_HFLIP:
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 0);
//	case V4L2_CID_GAIN:
//		return v4l2_ctrl_query_fill(qc, 0, 255, 1, 128);
//	case V4L2_CID_AUTOGAIN:
//		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 1);
	case V4L2_CID_EXPOSURE:
		return v4l2_ctrl_query_fill(qc, -4, 4, 1, 0);
	case V4L2_CID_EXPOSURE_AUTO:
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 0);
	case V4L2_CID_DO_WHITE_BALANCE:
		return v4l2_ctrl_query_fill(qc, 0, 5, 1, 0);
	case V4L2_CID_AUTO_WHITE_BALANCE:
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 1);
	case V4L2_CID_COLORFX:
		return v4l2_ctrl_query_fill(qc, 0, 9, 1, 0);
	case V4L2_CID_CAMERA_FLASH_MODE:
	  return v4l2_ctrl_query_fill(qc, 0, 4, 1, 0);
	case V4L2_CID_CAMERA_AE_TARGET:
		printk("query V4L2_CID_CAMERA_AE_TARGET\n");
		return v4l2_ctrl_query_fill(qc, 0, 1, 1, 0);
	}
	return -EINVAL;
}

static int sensor_g_hflip(struct v4l2_subdev *sd, __s32 *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x22;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_hflip!\n");
		return ret;
	}
	
	regs.value[0] &= (1<<1);
	regs.value[0] = regs.value[0]>>1;
		
	*value = regs.value[0];

	info->hflip = *value;
	return 0;
}

static int sensor_s_hflip(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;

	
	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x22;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_hflip!\n");
		return ret;
	}

	
	switch (value) {
		case 0:
		  regs.value[0] &= 0xFD;   //hflip is bit1 --> 0b 1111 1101
			break;
		case 1:
			regs.value[0] |= 0x02;
			break;
		default:
			return -EINVAL;
	}
	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_hflip!\n");
		return ret;
	}
	
	
	info->hflip = value;
	return 0;
}

static int sensor_g_vflip(struct v4l2_subdev *sd, __s32 *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;

	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x22;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_vflip!\n");
		return ret;
	}
	
	regs.value[0] &= (1<<0);
	regs.value[0] = regs.value[0]>>0;
		
	*value = regs.value[0];

	info->vflip = *value;
	return 0;
}

static int sensor_s_vflip(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x22;

	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_vflip!\n");
		return ret;
	}
	
	switch (value) {
		case 0:
		  regs.value[0] &= 0xFE;  //vflip is bit0 --> 0b 1111 1110
			break;
		case 1:
			regs.value[0] |= 0x01;
			break;
		default:
			return -EINVAL;
	}
	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_vflip!\n");
		return ret;
	}
	
	
	info->vflip = value;
	return 0;
}

static int sensor_g_autogain(struct v4l2_subdev *sd, __s32 *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;

	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0x00;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_autogain!\n");
		return ret;
	}
	
	regs.value[0] &= (1<<3);
	regs.value[0] = regs.value[0]>>3;
		
	*value = regs.value[0];

	info->autogain = *value;
	return 0;
}

static int sensor_s_autogain(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0x00;

	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_autogain!\n");
		return ret;
	}
	
	switch (value) {
		case 0:
		  regs.value[0] &= 0xF7;
			break;
		case 1:
			regs.value[0] |= 0x08;
			break;
		default:
			return -EINVAL;
	}
	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_autogain!\n");
		return ret;
	}
	
	
	info->autogain = value;
	return 0;
}

static int sensor_g_autoexp(struct v4l2_subdev *sd, __s32 *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0x01;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_autoexp!\n");
		return ret;
	}

	regs.value[0] &= 0x20;
	if (regs.value[0] == 0x20) {
		*value = V4L2_EXPOSURE_AUTO;
	}
	else
	{
		*value = V4L2_EXPOSURE_MANUAL;
	}
	
	info->autoexp = *value;
	return 0;
}

static int sensor_s_autoexp(struct v4l2_subdev *sd,
		enum v4l2_exposure_auto_type value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0x01;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_autoexp!\n");
		return ret;
	}

	switch (value) {
		case V4L2_EXPOSURE_AUTO:
		  regs.value[0] |= 0x20;
			break;
		case V4L2_EXPOSURE_MANUAL:
			regs.value[0] &= 0xDf;
			break;
		case V4L2_EXPOSURE_SHUTTER_PRIORITY:
			return -EINVAL;    
		case V4L2_EXPOSURE_APERTURE_PRIORITY:
			return -EINVAL;
		default:
			return -EINVAL;
	}
		
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_autoexp!\n");
		return ret;
	}
	
	
	info->autoexp = value;
	return 0;
}

static int sensor_g_autowb(struct v4l2_subdev *sd, int *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0x01;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_autowb!\n");
		return ret;
	}

	regs.value[0] &= (1<<4);
	regs.value[0] = regs.value[0]>>4;
		
	*value = regs.value[0];
	info->autowb = *value;
	
	return 0;
}

static int sensor_s_autowb(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0x01;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_autowb!\n");
		return ret;
	}

	switch(value) {
	case 0:
		regs.value[0] &= 0xEf;
		break;
	case 1:
		regs.value[0] |= 0x10;
		break;
	default:
		break;
	}	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_autowb!\n");
		return ret;
	}
	
	
	info->autowb = value;
	return 0;
}

static int sensor_g_hue(struct v4l2_subdev *sd, __s32 *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0xF1;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_hue!\n");
		return ret;
	}

	*value = ((regs.value[0] & 0x05)==0x05) ? 1: 0;
	info->hue = *value;
	
	return 0;
}

static int sensor_s_hue(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0xF1;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_hue!\n");
		return ret;
	}

	switch(value) {
	case 0:
		regs.value[0] &= 0xF8;
		break;
	case 1:
		regs.value[0] |= 0x05;
		break;
	default:
		break;
	}	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_hue!\n");
		return ret;
	}
	
	
	info->hue = value;
	return 0;
}

static int sensor_g_gain(struct v4l2_subdev *sd, __s32 *value)
{
	return -EINVAL;
}

static int sensor_s_gain(struct v4l2_subdev *sd, int value)
{
	return -EINVAL;
}
/* *********************************************end of ******************************************** */

static int sensor_g_brightness(struct v4l2_subdev *sd, __s32 *value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0xFC;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_g_brightness!\n");
		return ret;
	}

	*value = regs.value[0];
	info->brightness = *value;
	
	return 0;
}

static int sensor_s_brightness(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	struct regval_list regs;
	
	regs.reg_num[0] = 0x32;
	regs.reg_num[1] = 0xFC;
	ret = sensor_read(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_read err at sensor_s_brightness 1!\n");
		return ret;
	}
	csi_dev_dbg("ensor_s_brightness at 1 value %d\n", value);
	if(value>=-4 && value<=4)
	{
		value = value<<3;
		if(value>0)
			value--;
	}

	csi_dev_dbg("sensor_s_brightness at 2 value %d\n", value);
	regs.value[0] = (unsigned char)value;
	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_brightness 2!\n");
		return ret;
	}

	/* active */
	regs.reg_num[0] = 0x30;
	regs.reg_num[1] = 0x60;
	regs.value[0] = (unsigned char)0x01;
	
	ret = sensor_write(sd, regs.reg_num, regs.value);
	if (ret < 0) {
		csi_dev_err("sensor_write err at sensor_s_brightness 3!\n");
		return ret;
	}
	
	info->brightness = value;
	return 0;
}

static int sensor_g_contrast(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->contrast;
	return 0;
}

static int sensor_s_contrast(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_contrast_neg4_regs, ARRAY_SIZE(sensor_contrast_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_contrast_neg3_regs, ARRAY_SIZE(sensor_contrast_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_contrast_neg2_regs, ARRAY_SIZE(sensor_contrast_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_contrast_neg1_regs, ARRAY_SIZE(sensor_contrast_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_contrast_zero_regs, ARRAY_SIZE(sensor_contrast_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_contrast_pos1_regs, ARRAY_SIZE(sensor_contrast_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_contrast_pos2_regs, ARRAY_SIZE(sensor_contrast_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_contrast_pos3_regs, ARRAY_SIZE(sensor_contrast_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_contrast_pos4_regs, ARRAY_SIZE(sensor_contrast_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_contrast!\n");
		return ret;
	}
	
	
	info->contrast = value;
	return 0;
}

static int sensor_g_saturation(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->saturation;
	return 0;
}

static int sensor_s_saturation(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);
	
	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_saturation_neg4_regs, ARRAY_SIZE(sensor_saturation_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_saturation_neg3_regs, ARRAY_SIZE(sensor_saturation_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_saturation_neg2_regs, ARRAY_SIZE(sensor_saturation_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_saturation_neg1_regs, ARRAY_SIZE(sensor_saturation_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_saturation_zero_regs, ARRAY_SIZE(sensor_saturation_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_saturation_pos1_regs, ARRAY_SIZE(sensor_saturation_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_saturation_pos2_regs, ARRAY_SIZE(sensor_saturation_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_saturation_pos3_regs, ARRAY_SIZE(sensor_saturation_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_saturation_pos4_regs, ARRAY_SIZE(sensor_saturation_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_saturation!\n");
		return ret;
	}
	
	
	info->saturation = value;
	return 0;
}

static int sensor_g_exp(struct v4l2_subdev *sd, __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	
	*value = info->exp;
	return 0;
}

static int sensor_s_exp(struct v4l2_subdev *sd, int value)
{
	int ret;
	struct sensor_info *info = to_state(sd);

	switch (value) {
		case -4:
		  ret = sensor_write_array(sd, sensor_ev_neg4_regs, ARRAY_SIZE(sensor_ev_neg4_regs));
			break;
		case -3:
			ret = sensor_write_array(sd, sensor_ev_neg3_regs, ARRAY_SIZE(sensor_ev_neg3_regs));
			break;
		case -2:
			ret = sensor_write_array(sd, sensor_ev_neg2_regs, ARRAY_SIZE(sensor_ev_neg2_regs));
			break;   
		case -1:
			ret = sensor_write_array(sd, sensor_ev_neg1_regs, ARRAY_SIZE(sensor_ev_neg1_regs));
			break;
		case 0:   
			ret = sensor_write_array(sd, sensor_ev_zero_regs, ARRAY_SIZE(sensor_ev_zero_regs));
			break;
		case 1:
			ret = sensor_write_array(sd, sensor_ev_pos1_regs, ARRAY_SIZE(sensor_ev_pos1_regs));
			break;
		case 2:
			ret = sensor_write_array(sd, sensor_ev_pos2_regs, ARRAY_SIZE(sensor_ev_pos2_regs));
			break;	
		case 3:
			ret = sensor_write_array(sd, sensor_ev_pos3_regs, ARRAY_SIZE(sensor_ev_pos3_regs));
			break;
		case 4:
			ret = sensor_write_array(sd, sensor_ev_pos4_regs, ARRAY_SIZE(sensor_ev_pos4_regs));
			break;
		default:
			return -EINVAL;
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_exp!\n");
		return ret;
	}
	
	
	info->exp = value;
	return 0;
}

static int sensor_g_wb(struct v4l2_subdev *sd, int *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_whiteblance *wb_type = (enum v4l2_whiteblance*)value;
	
	*wb_type = info->wb;
	
	return 0;
}

static int sensor_s_wb(struct v4l2_subdev *sd,
		enum v4l2_whiteblance value)
{
	int ret;
	struct sensor_info *info = to_state(sd);

	csi_dev_err("sensor_s_wb value %d\n",value);
	if (value == V4L2_WB_AUTO) {
		ret = sensor_s_autowb(sd, 1);
		return ret;
	} 
	else {
		ret = sensor_s_autowb(sd, 0);
		if(ret < 0) {
			csi_dev_err("sensor_s_autowb error, return %x!\n",ret);
			return ret;
		}
		
		switch (value) {
			case V4L2_WB_CLOUD:
			  ret = sensor_write_array(sd, sensor_wb_cloud_regs, ARRAY_SIZE(sensor_wb_cloud_regs));
				break;
			case V4L2_WB_DAYLIGHT:
				ret = sensor_write_array(sd, sensor_wb_daylight_regs, ARRAY_SIZE(sensor_wb_daylight_regs));
				break;
			case V4L2_WB_INCANDESCENCE:
				ret = sensor_write_array(sd, sensor_wb_incandescence_regs, ARRAY_SIZE(sensor_wb_incandescence_regs));
				break;    
			case V4L2_WB_FLUORESCENT:
				ret = sensor_write_array(sd, sensor_wb_fluorescent_regs, ARRAY_SIZE(sensor_wb_fluorescent_regs));
				break;
			case V4L2_WB_TUNGSTEN:   
				ret = sensor_write_array(sd, sensor_wb_tungsten_regs, ARRAY_SIZE(sensor_wb_tungsten_regs));
				break;
			default:
				return -EINVAL;
		} 
	}
	
	if (ret < 0) {
		csi_dev_err("sensor_s_wb error, return %x!\n",ret);
		return ret;
	}
	
	
	info->wb = value;
	return 0;
}

static int sensor_g_colorfx(struct v4l2_subdev *sd,
		__s32 *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_colorfx *clrfx_type = (enum v4l2_colorfx*)value;
	
	*clrfx_type = info->clrfx;
	return 0;
}

static int sensor_s_colorfx(struct v4l2_subdev *sd,
		enum v4l2_colorfx value)
{
	int ret;

	struct sensor_info *info = to_state(sd);
	switch (value) {
	case V4L2_COLORFX_NONE:
	  ret = sensor_write_array(sd, sensor_colorfx_none_regs, ARRAY_SIZE(sensor_colorfx_none_regs));
		break;
	case V4L2_COLORFX_BW:
		ret = sensor_write_array(sd, sensor_colorfx_bw_regs, ARRAY_SIZE(sensor_colorfx_bw_regs));
		break;  
	case V4L2_COLORFX_SEPIA:
		ret = sensor_write_array(sd, sensor_colorfx_sepia_regs, ARRAY_SIZE(sensor_colorfx_sepia_regs));
		break;   
	case V4L2_COLORFX_NEGATIVE:
		ret = sensor_write_array(sd, sensor_colorfx_negative_regs, ARRAY_SIZE(sensor_colorfx_negative_regs));
		break;
	case V4L2_COLORFX_EMBOSS:   
		ret = sensor_write_array(sd, sensor_colorfx_emboss_regs, ARRAY_SIZE(sensor_colorfx_emboss_regs));
		break;
	case V4L2_COLORFX_SKETCH:     
		ret = sensor_write_array(sd, sensor_colorfx_sketch_regs, ARRAY_SIZE(sensor_colorfx_sketch_regs));
		break;
	case V4L2_COLORFX_SKY_BLUE:
		ret = sensor_write_array(sd, sensor_colorfx_sky_blue_regs, ARRAY_SIZE(sensor_colorfx_sky_blue_regs));
		break;
	case V4L2_COLORFX_GRASS_GREEN:
		ret = sensor_write_array(sd, sensor_colorfx_grass_green_regs, ARRAY_SIZE(sensor_colorfx_grass_green_regs));
		break;
	case V4L2_COLORFX_SKIN_WHITEN:
		ret = sensor_write_array(sd, sensor_colorfx_skin_whiten_regs, ARRAY_SIZE(sensor_colorfx_skin_whiten_regs));
		break;
	case V4L2_COLORFX_VIVID:
		ret = sensor_write_array(sd, sensor_colorfx_vivid_regs, ARRAY_SIZE(sensor_colorfx_vivid_regs));
		break;
	default:
		return -EINVAL;
	}

	if (ret < 0) {
		csi_dev_err("sensor_s_colorfx error, return %x!\n",ret);
		return ret;
	}
	
	
	info->clrfx = value;
	return 0;
}

static int sensor_g_flash_mode(struct v4l2_subdev *sd,
    __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_flash_mode *flash_mode = (enum v4l2_flash_mode*)value;
	
	*flash_mode = info->flash_mode;
	return 0;
}

static int sensor_s_flash_mode(struct v4l2_subdev *sd,
    enum v4l2_flash_mode value)
{
	struct sensor_info *info = to_state(sd);
	struct csi_dev *dev=(struct csi_dev *)dev_get_drvdata(sd->v4l2_dev->dev);
	int flash_on,flash_off;
	
	flash_on = (dev->flash_pol!=0)?1:0;
	flash_off = (flash_on==1)?0:1;
	
	switch (value) {
	case V4L2_FLASH_MODE_OFF:
		csi_gpio_write(sd,&dev->flash_io,flash_off);
		break;
	case V4L2_FLASH_MODE_AUTO:
		return -EINVAL;
		break;  
	case V4L2_FLASH_MODE_ON:
		csi_gpio_write(sd,&dev->flash_io,flash_on);
		break;   
	case V4L2_FLASH_MODE_TORCH:
		return -EINVAL;
		break;
	case V4L2_FLASH_MODE_RED_EYE:   
		return -EINVAL;
		break;
	default:
		return -EINVAL;
	}
	
	info->flash_mode = value;
	return 0;
}

static int sensor_g_aetarget(struct v4l2_subdev *sd,
    __s32 *value)
{
	struct sensor_info *info = to_state(sd);
	enum v4l2_aetarget_ctrl *aetarget = (enum v4l2_aetarget_ctrl*)value;
	
	*aetarget = info->aetarget;
	return 0;
}

static int sensor_s_aetarget(struct v4l2_subdev *sd,
    enum v4l2_aetarget_ctrl value)
{
	int ret;
	struct sensor_info *info = to_state(sd);

	printk("sensor_s_aetarget value is %d\n", value);
	
	switch (value) {
	case V4L2_AETARGET_0://hangzhou brightness
		ret = sensor_write_array(sd, sensor_ae_target0_default_regs, ARRAY_SIZE(sensor_ae_target0_default_regs));		
		break;
	case V4L2_AETARGET_1:
		ret = sensor_write_array(sd, sensor_ae_target1_default_regs, ARRAY_SIZE(sensor_ae_target1_default_regs));
		break;
	default:
		return -EINVAL;
	}

	if (ret < 0) {
		csi_dev_err("sensor_write_array err at sensor_s_aetarget!\n");
		return ret;
	}

	info->aetarget = value;
	return 0;
}


static int sensor_g_ctrl(struct v4l2_subdev *sd, struct v4l2_control *ctrl)
{
	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		return sensor_g_brightness(sd, &ctrl->value);
	case V4L2_CID_CONTRAST:
		return sensor_g_contrast(sd, &ctrl->value);
	case V4L2_CID_SATURATION:
		return sensor_g_saturation(sd, &ctrl->value);
	case V4L2_CID_HUE:
		return sensor_g_hue(sd, &ctrl->value);	
	case V4L2_CID_VFLIP:
		return sensor_g_vflip(sd, &ctrl->value);
	case V4L2_CID_HFLIP:
		return sensor_g_hflip(sd, &ctrl->value);
	case V4L2_CID_GAIN:
		return sensor_g_gain(sd, &ctrl->value);
	case V4L2_CID_AUTOGAIN:
		return sensor_g_autogain(sd, &ctrl->value);
	case V4L2_CID_EXPOSURE:
		return sensor_g_exp(sd, &ctrl->value);
	case V4L2_CID_EXPOSURE_AUTO:
		return sensor_g_autoexp(sd, &ctrl->value);
	case V4L2_CID_DO_WHITE_BALANCE:
		return sensor_g_wb(sd, &ctrl->value);
	case V4L2_CID_AUTO_WHITE_BALANCE:
		return sensor_g_autowb(sd, &ctrl->value);
	case V4L2_CID_COLORFX:
		return sensor_g_colorfx(sd,	&ctrl->value);
	case V4L2_CID_CAMERA_FLASH_MODE:
		return sensor_g_flash_mode(sd, &ctrl->value);
	case V4L2_CID_CAMERA_AE_TARGET:
		return sensor_g_aetarget(sd, &ctrl->value);
	}
	return -EINVAL;
}

static int sensor_s_ctrl(struct v4l2_subdev *sd, struct v4l2_control *ctrl)
{

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		return sensor_s_brightness(sd, ctrl->value);
	case V4L2_CID_CONTRAST:
		return sensor_s_contrast(sd, ctrl->value);
	case V4L2_CID_SATURATION:
		return sensor_s_saturation(sd, ctrl->value);
	case V4L2_CID_HUE:
		return sensor_s_hue(sd, ctrl->value);		
	case V4L2_CID_VFLIP:
		return sensor_s_vflip(sd, ctrl->value);
	case V4L2_CID_HFLIP:
		return sensor_s_hflip(sd, ctrl->value);
	case V4L2_CID_GAIN:
		return sensor_s_gain(sd, ctrl->value);
	case V4L2_CID_AUTOGAIN:
		return sensor_s_autogain(sd, ctrl->value);
	case V4L2_CID_EXPOSURE:
		return sensor_s_exp(sd, ctrl->value);
	case V4L2_CID_EXPOSURE_AUTO:
		return sensor_s_autoexp(sd,
				(enum v4l2_exposure_auto_type) ctrl->value);
	case V4L2_CID_DO_WHITE_BALANCE:
		return sensor_s_wb(sd,
				(enum v4l2_whiteblance) ctrl->value);	
	case V4L2_CID_AUTO_WHITE_BALANCE:
		return sensor_s_autowb(sd, ctrl->value);
	case V4L2_CID_COLORFX:
		return sensor_s_colorfx(sd,
				(enum v4l2_colorfx) ctrl->value);
	case V4L2_CID_CAMERA_FLASH_MODE:
	  return sensor_s_flash_mode(sd,
	      (enum v4l2_flash_mode) ctrl->value);
	case V4L2_CID_CAMERA_AE_TARGET:
		return sensor_s_aetarget(sd, (enum v4l2_aetarget_ctrl)ctrl->value);
	}
	return -EINVAL;
}

static int sensor_g_chip_ident(struct v4l2_subdev *sd,
		struct v4l2_dbg_chip_ident *chip)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	return v4l2_chip_ident_i2c_client(client, chip, V4L2_IDENT_SENSOR, 0);
}

static int hexchar2int(char* buf, int start, int len) 
{ 
	int sum = 0, b = 0; 
       int i;
	char* a = buf + start;
	for( i=0;i<len;i++)
	{
		sum = sum*16;
		if (a[i]>='0' && a[i]<='9')
			b=a[i]-'0'; 
		else if(a[i]>='a' && a[i]<='f')
			b=a[i]-'a'+10; 
		else if(a[i]>='A' && a[i]<='F')
			b=a[i]-'A'+10; 
		else
			b = 0;
			
		sum += b;
		//printk("hexchar2int b is %d, sum is %d\n ", b, sum );
	}

	return sum;
}

static ssize_t write_i2c(struct device *dev,
                          struct device_attribute *attr, const char *buf, size_t size)
{
    struct i2c_client *client =  i2c_verify_client(dev);
    unsigned char reg[REG_ADDR_STEP], value[REG_DATA_STEP];
    char desc_format[]="%hx %hx\n";
    int ret;

    /*CMD:       ADDR    VALUE*
     *for example: echo "0x01,0xff" > /sys/...
     * write value 0xff to page 0 off= 0x01*
     */

    if(NULL == client)
        printk("failed to get i2c_client\n");

	if (!strncmp(buf, "30fps", strlen("30fps")))
	{
		printk(KERN_ERR "set to 30fps\n");
		return sensor_write_array_client(client, sensor_ae_target0_hd720_30fps_regs , ARRAY_SIZE(sensor_ae_target0_hd720_30fps_regs));
	}
	else if (!strncmp(buf, "25fps", strlen("25fps")))
	{
		printk(KERN_ERR "set to 25fps\n");
		return sensor_write_array_client(client, sensor_ae_target0_hd720_25fps_regs , ARRAY_SIZE(sensor_ae_target0_hd720_25fps_regs));
	}
	else if(!strncmp(buf, "29fps", strlen("29fps")))
	{
		printk(KERN_ERR "set to 29fps\n");
		return sensor_write_array_client(client, sensor_ae_target0_hd720_29fps_regs , ARRAY_SIZE(sensor_ae_target0_hd720_29fps_regs));  
      }
	else if (!strncmp(buf, "15fps", strlen("15fps")))
	{
		printk(KERN_ERR "set to 15fps\n");
		return sensor_write_array_client(client, sensor_ae_target0_hd720_15fps_regs , ARRAY_SIZE(sensor_ae_target0_hd720_15fps_regs));
	}
	else if (!strncmp(buf, "5fps", strlen("5fps")))
	{
		printk(KERN_ERR "set to 5fps\n");
		return sensor_write_array_client(client, sensor_ae_target0_hd720_5fps_regs , ARRAY_SIZE(sensor_ae_target0_hd720_5fps_regs));
	}	
	else
	{
	    //sscanf(buf, desc_format, reg, value);
	    ret = hexchar2int(buf, 0, REG_ADDR_STEP*2);
	    //printk("addr = %d\n", ret);
	    if(ret < 0)
		csi_dev_err("write_i2c buf err !\n");
		
	    reg[0] = ret >> 8;
	    reg[1] = (ret & 0xff);

	    ret = hexchar2int(buf, 5, 2);
	    value[0] = (ret & 0xff);
	    ret = sensor_write_client(client, reg, value);
	    if(ret < 0)
	        printk("i2c write failed\n");
	}
 
	return size;
}

static ssize_t read_i2c(struct device *dev,
                          struct device_attribute *attr, const char *buf, size_t size)
{
	int ret;
    struct i2c_client *client =  i2c_verify_client(dev);
    unsigned char reg[REG_ADDR_STEP], value[REG_DATA_STEP];
    char desc_format[]="%hx\n";

    if(NULL == client)
        printk("failed to get i2c_client\n");

    //sscanf(buf,desc_format, reg);
    printk("read all reg %s\n", buf);
    if (!strncmp(buf, "readall", strlen("readall")))
    {
             for(reg[0]=0x30;reg[0]<0x34;reg[0]++)
             	{
	             	for(reg[1]=0x00;reg[1]<0xFF;)
	             	{
				ret = sensor_read_client(client, reg, value);
				if (ret >= 0)
				{
					pr_alert("reg:0x%x%x, 0x%x\n", reg[0], reg[1], value[0]);
					reg[1]++;
				}
	             	}
             	}
		printk("finish read all reg:\n");
		return size;
    }

    ret = hexchar2int(buf, 0, REG_ADDR_STEP*2);
    if(ret < 0)
	csi_dev_err("sensor_read buf err !\n");
	
    reg[0] = ret >> 8;
    reg[1] = (ret & 0xff);
	
    //printk("read i2c reg %x,%x, addr is %x\n", reg[0], reg[1], client->addr);
    ret = sensor_read_client(client, reg, value);
	if (ret < 0)
	{
		csi_dev_err("sensor_read err!\n");		
	}
	else
	{
		pr_alert("reg:0x%x%x, 0x%x\n", reg[0], reg[1], value[0]);
	}
	
	return size;
}

static DEVICE_ATTR(i2cw, S_IRUGO|S_IWUGO, NULL, write_i2c);
static DEVICE_ATTR(i2cr, S_IRUGO|S_IWUSR, NULL, read_i2c);
static struct device_attribute *nt99141_sysfs_attrs[] = {
    &dev_attr_i2cw,
    &dev_attr_i2cr,
    NULL,
};

static int nt99141_sys_init(struct i2c_client *client)
{
    struct device_attribute *attr;
    int retval;
    int i = 0;

    while ((attr = nt99141_sysfs_attrs[i++]) != NULL) {
      retval=device_create_file (&(client->dev),attr);
      if (retval != 0) {
         goto out_create_file;
      }
   }

    /* create display sysfs links */
   retval = sysfs_create_link(NULL,&(client->dev.kobj), DRIVER_NAME);
   if (retval != 0)
     goto out_create_link;

   printk("client addr = %x\n", client->addr);
   return retval;

 out_create_link:
   sysfs_remove_link(NULL, DRIVER_NAME);
 out_create_file:
   while ((attr = nt99141_sysfs_attrs[i++]) != NULL) {
      device_remove_file (&(client->dev),attr);
   }
   return retval;
}

/* ----------------------------------------------------------------------- */

static const struct v4l2_subdev_core_ops sensor_core_ops = {
	.g_chip_ident = sensor_g_chip_ident,
	.g_ctrl = sensor_g_ctrl,
	.s_ctrl = sensor_s_ctrl,
	.queryctrl = sensor_queryctrl,
	.reset = sensor_reset,
	.init = sensor_init,
	.s_power = sensor_power,
	.ioctl = sensor_ioctl,
};

static const struct v4l2_subdev_video_ops sensor_video_ops = {
	.enum_mbus_fmt = sensor_enum_fmt,//linux-3.0
	.try_mbus_fmt = sensor_try_fmt,//linux-3.0
	.s_mbus_fmt = sensor_s_fmt,//linux-3.0
	.s_parm = sensor_s_parm,//linux-3.0
	.g_parm = sensor_g_parm,//linux-3.0
};

static const struct v4l2_subdev_ops sensor_ops = {
	.core = &sensor_core_ops,
	.video = &sensor_video_ops,
};

/* ----------------------------------------------------------------------- */

static int sensor_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct v4l2_subdev *sd;
	struct sensor_info *info;
//	int ret;

	info = kzalloc(sizeof(struct sensor_info), GFP_KERNEL);
	if (info == NULL)
		return -ENOMEM;
	sd = &info->sd;
	v4l2_i2c_subdev_init(sd, client, &sensor_ops);

	info->fmt = &sensor_formats[0];
	info->ccm_info = &ccm_info_con;

	info->brightness = 0;
	info->contrast = 0;
	info->saturation = 0;
	info->hue = 0;
	info->hflip = 0;
	info->vflip = 0;
	info->gain = 0;
	info->autogain = 1;
	info->exp = 0;
	info->autoexp = 0;
	info->autowb = 1;
	info->wb = 0;
	info->clrfx = 0;
	info->aetarget = 0;
	
//	info->clkrc = 1;	/* 30fps */
	nt99141_sys_init(client);

	return 0;
}


static int sensor_remove(struct i2c_client *client)
{
	struct v4l2_subdev *sd = i2c_get_clientdata(client);

	v4l2_device_unregister_subdev(sd);
	kfree(to_state(sd));
	return 0;
}

static const struct i2c_device_id sensor_id[] = {
	{ "nt99141", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, sensor_id);

//linux-3.0
static struct i2c_driver sensor_driver = {
	.driver = {
		.owner = THIS_MODULE,
	.name = "nt99141",
	},
	.probe = sensor_probe,
	.remove = sensor_remove,
	.id_table = sensor_id,
};
static __init int init_sensor(void)
{
	return i2c_add_driver(&sensor_driver);
}

static __exit void exit_sensor(void)
{
  i2c_del_driver(&sensor_driver);
}

module_init(init_sensor);
module_exit(exit_sensor);


